# -*- coding: utf-8 -*-
"""
Various utilies related to logging, saving and loading files, edge detection
and sampling latent distributions.
"""
import json
import logging
import os
import shutil

from nflows.distributions.uniform import BoxUniform
import numpy as np
from scipy import stats, spatial
import torch
from torch.distributions import MultivariateNormal

from .livepoint import live_points_to_dict

logger = logging.getLogger(__name__)


def logit(x, fuzz=1e-2):
    """
    Logit function that also returns log Jacobian

    Parameters
    ----------
    x : ndarray
        Array of values
    fuzz : float, optional
        Fuzz used to avoid nans in logit. Values are rescaled from [0, 1]
        to [0-fuzz, 1+fuzz]. By default no fuzz is applied
    """
    x += fuzz
    x /= (1 + 2 * fuzz)
    return np.log(x) - np.log(1 - x), -np.log(np.abs(x - x ** 2))


def sigmoid(x, fuzz=1e-2):
    """
    Sigmoid function that also returns log Jacobian

    Parameters
    ----------
    x : ndarray
        Array of values
    fuzz : float, optional
        Fuzz used to avoid nans in logit
    """
    x = np.asarray(x)
    x = np.divide(1, 1 + np.exp(-x))
    log_J = np.log(np.abs(x - x ** 2))
    x *= (1 + 2 * fuzz)
    x -= fuzz
    return x, log_J


rescaling_functions = {'logit': (logit, sigmoid)}


def compute_indices_ks_test(indices, nlive, mode='D+'):
    """
    Compute the two-sided KS test for discrete insertion indices for a given
    number of live points

    Parameters
    ----------
    indices : array_like
        Indices of newly inserteed live points
    nlive : int
        Number of live points

    Returns
    ------
    D : float
        Two-sided KS statistic
    p : float
        p-value
    """
    if len(indices):
        counts = np.zeros(nlive)
        u, c = np.unique(indices, return_counts=True)
        counts[u] = c
        cdf = np.cumsum(counts) / len(indices)
        if mode == 'D+':
            D = np.max(np.arange(1.0, nlive + 1) / nlive - cdf)
        elif mode == 'D-':
            D = np.max(cdf - np.arange(0.0, nlive) / nlive)
        else:
            raise RuntimeError(f'{mode} is not a valid mode. Choose D+ or D-')
        p = stats.ksone.sf(D, len(indices))
        return D, p
    else:
        return None, None


def bonferroni_correction(p_values, alpha=0.05):
    """
    Apply the Bonferroni correction for multiple tests.

    Based on the implementation in ``statmodels.stats.multitest``

    Parameters
    ----------
    p_values :  array_like
        Uncorrelated p-values.
    alpha : float, optional
        Family wise error rate.
    """
    p_values = np.asarray(p_values)
    alpha_bon = alpha / p_values.size
    reject = p_values <= alpha_bon
    p_values_corrected = p_values * p_values.size
    p_values_corrected[p_values_corrected > 1] = 1
    return reject, p_values_corrected, alpha_bon


def draw_surface_nsphere(dims, r=1, N=1000):
    """
    Draw N points uniformly from  n-1 sphere of radius r using Marsaglia's
    algorithm. E.g for 3 dimensions returns points on a 'regular' sphere.

    See Marsaglia (1972)

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-sphere, if specified it is used to rescale the samples
    N : int, optional
        Number of samples to draw

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = np.random.randn(N, dims)
    R = np.sqrt(np.sum(x ** 2., axis=1))[:, np.newaxis]
    z = x / R
    return r * z


def draw_nsphere(dims, r=1, N=1000, fuzz=1.0):
    """
    Draw N points uniformly within an n-sphere of radius r

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, optional
        Number of samples to draw
    fuzz : float, optional
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    x = draw_surface_nsphere(dims, r=1, N=N)
    R = np.random.uniform(0, 1, (N, 1))
    z = R ** (1 / dims) * x
    return fuzz * r * z


def get_uniform_distribution(dims, r, device='cpu'):
    """
    Return a torch distribution that is uniform in the number of dims
    specified.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    r : float
        Radius to use for lower and upper bounds.
    device : str, optional
        Device on which the distribution is placed.

    Returns
    -------
    :obj:`nflows.distributions.uniform.BoxUniform`
        Instance of BoxUniform which the lower and upper bounds set by
        the radius
    """
    r = r * torch.ones(dims, device=device)
    return BoxUniform(low=-r, high=r)


def get_multivariate_normal(dims, var=1, device='cpu'):
    """
    Return a Pytorch distribution that is normally distributed in n dims
    with a given variance.

    Parameters
    ----------
    dims : int
        Number of dimensions.
    var : float, optional
        Variance.
    device : str, optional
        Device on which the distribution is placed.

    Returns
    -------
    :obj:`nessai.flows.distributions.MultivariateNormal`
        Instance of MultivariateNormal with correct variance and dims.
    """
    loc = torch.zeros(dims).to(device).double()
    covar = var * torch.eye(dims).to(device).double()
    return MultivariateNormal(loc, covariance_matrix=covar)


def draw_uniform(dims, r=(1,), N=1000, fuzz=1.0):
    """
    Draw from a uniform distribution on [0, 1].

    Deals with extra input parameters used by other draw functions

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball. (Ignored by this function)
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball. (Ingored by
        this function)

    Returns
    -------
    ndarraay
        Array of samples with shape (N, dims)
    """
    return np.random.uniform(0, 1, (N, dims))


def draw_gaussian(dims, r=1, N=1000, fuzz=1.0):
    """
    Wrapper for numpy.random.randn that deals with extra input parameters
    r and fuzz

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float, optional
        Radius of the n-ball
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the n-ball

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    return np.random.randn(N, dims)


def draw_truncated_gaussian(dims, r, N=1000, fuzz=1.0, var=1):
    """
    Draw N points from a truncated gaussian with a given a radius

    Parameters
    ----------
    dims : int
        Dimension of the n-sphere
    r : float
        Radius of the truncated Gaussian
    N : int, ignored
        Number of samples to draw
    fuzz : float, ignored
        Fuzz factor by which to increase the radius of the truncated Gaussian

    Returns
    -------
    ndarray
        Array of samples with shape (N, dims)
    """
    sigma = np.sqrt(var)
    r *= fuzz
    u_max = stats.chi.cdf(r / sigma, df=dims)
    u = np.random.uniform(0, u_max, N)
    p = sigma * stats.chi.ppf(u, df=dims)
    x = np.random.randn(p.size, dims)
    points = (p * x.T / np.sqrt(np.sum(x**2., axis=1))).T
    return points


def replace_in_list(target_list, targets, replacements):
    """
    Replace (in place) an entry in a list with a given element

    Parameters
    ----------
    target_list : list
        List to update
    targets : list
        List of items to update
    replacements : list
        List of replacement items
    """
    if not isinstance(targets, list):
        if isinstance(targets, int):
            targets = [targets]
        else:
            targets = list(targets)
    if not isinstance(replacements, list):
        if isinstance(replacements, int):
            replacements = [replacements]
        else:
            replacements = list(replacements)

    if not all([t in target_list for t in targets]):
        raise ValueError(f'Targets {targets} not in list: {target_list}')

    for t, r in zip(targets, replacements):
        i = target_list.index(t)
        target_list[i] = r


def rescale_zero_to_one(x, xmin, xmax):
    """
    Rescale a value to 0 to 1

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (x - xmin) / (xmax - xmin), -np.log(xmax - xmin)


def inverse_rescale_zero_to_one(x, xmin, xmax):
    """
    Rescale from 0 to 1 to xmin to xmax

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (xmax - xmin) * x + xmin, np.log(xmax - xmin)


def rescale_minus_one_to_one(x, xmin, xmax):
    """
    Rescale a value to -1 to 1

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return ((2. * (x - xmin) / (xmax - xmin)) - 1,
            np.log(2) - np.log(xmax - xmin))


def inverse_rescale_minus_one_to_one(x, xmin, xmax):
    """
    Rescale from -1 to 1 to xmin to xmax

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return ((xmax - xmin) * ((x + 1) / 2.) + xmin,
            np.log(xmax - xmin) - np.log(2))


def detect_edge(x, x_range=None, percent=0.1, cutoff=0.5, nbins='auto',
                allow_both=False, allow_none=False,
                allowed_bounds=['lower', 'upper'], test=None):
    """
    Detect edges in input distributions based on the density.

    Parameters
    ----------
    x: array_like
        Samples
    x_range : array_like, optional
        Lower and upper bounds used to check inversion, if not specified
        min and max of data are used.
    percent: float (0.1)
        Percentage of interval used to check edges
    cutoff: float (0.1)
        Minimum fraction of the maximum density contained within the
        percentage of the interval specified
    nbins : float or 'auto'
        Number of bins used for histogram.
    allow_both: bool
        Allow function to return both instead of force either upper or lower
    allow_none: bool
        Allow for neither lower or upper bound to be returned
    allowed_bounds : list
        List of alloweds bounds.
    test : str or None
        If not None this skips the process and just returns the value of test.
        This is used to verify the inversion in all possible scenarios.

    Returns
    -------
    str or False, {'lower', 'upper', 'both', False}
        Returns the boundary to apply the inversion or False is no inversion
        is to be applied
    """
    bounds = ['lower', 'upper']
    if test is not None:
        logger.debug('Using test in detect_edge')
        if test in bounds and test not in allowed_bounds:
            logger.debug(f'{test} is not an allowed bound, returning False')
            return False
        else:
            return test
    if not all(b in bounds for b in allowed_bounds):
        raise RuntimeError(f'Unknown allowed bounds: {allowed_bounds}')
    if nbins == 'auto':
        nbins = auto_bins(x)

    hist, bins = np.histogram(x, bins=nbins, density=True, range=x_range)
    n = max(int(len(bins) * percent), 1)
    bounds_fraction = \
        np.array([np.sum(hist[:n]), np.sum(hist[-n:])]) * (bins[1] - bins[0])
    max_idx = np.argmax(hist)
    max_density = hist[max_idx] * (bins[1] - bins[0])
    logger.debug(f'Max. density: {max_density:.3f}')

    for i, b in enumerate(bounds):
        if b not in allowed_bounds:
            bounds.pop(i)
            bounds_fraction = np.delete(bounds_fraction, i)
    if max_idx <= n and 'lower' in bounds:
        return bounds[0]
    elif max_idx >= (len(bins) - n) and 'upper' in bounds:
        return bounds[-1]
    elif not np.any(bounds_fraction > cutoff * max_density) and allow_none:
        logger.debug('Density too low at both bounds')
        return False
    else:
        if (np.all(bounds_fraction > cutoff * max_density) and allow_both and
                len(bounds) > 1):
            logger.debug('Both bounds above cutoff')
            return 'both'
        else:
            return bounds[np.argmax(bounds_fraction)]


def configure_edge_detection(d, detect_edges):
    """
    Configure parameters for edge detection

    Parameters
    ----------
    d : dict
        Dictionary of kwargs passed to detect_edge.
    detect_edges : bool
        If true allows for no inversion to be applied.

    Returns
    -------
    dict
        Updated kwargs
    """
    default = dict(cutoff=0.5)
    if d is None:
        d = {}
    if detect_edges:
        d['allow_none'] = True
    else:
        d['allow_none'] = False
        d['cutoff'] = 0.0
    default.update(d)
    logger.debug(f'detect edges kwargs: {default}')
    return default


def compute_minimum_distances(samples, metric='euclidean'):
    """
    Compute the distance to the nearest neighbour of each sample

    Parameters
    ----------
    samples : array_like
        Array of samples.
    metric : str, optional
        Metric to use. See scipy docs for list of metrics:
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.cdist.html

    Returns
    -------
    array_like
        Distance to nearest neighbour for each sample.
    """
    d = spatial.distance.cdist(samples, samples, metric)
    d[d == 0] = np.nan
    dmin = np.nanmin(d, axis=1)
    return dmin


def setup_logger(output=None, label='nessai', log_level='INFO'):
    """
    Setup the logger.

    Based on the implementation in Bilby:
    https://git.ligo.org/lscsoft/bilby/-/blob/master/bilby/core/utils.py#L448

    Parameters
    ----------
    output : str, optional
        Path of to output directory.
    label : str, optional
        Label for this instance of the logger.
    log_level : {'ERROR', 'WARNING', 'INFO', 'DEBUG'}, optional
        Level of logging passed to logger.

    Returns
    -------
    :obj:`logging.Logger`
        Instance of the Logger class.
    """
    from . import __version__ as version
    if type(log_level) is str:
        try:
            level = getattr(logging, log_level.upper())
        except AttributeError:
            raise ValueError('log_level {} not understood'.format(log_level))
    else:
        level = int(log_level)

    logger = logging.getLogger('nessai')
    logger.propagate = False
    logger.setLevel(level)

    if any([type(h) == logging.StreamHandler for h in logger.handlers]) \
            is False:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(
            '%(asctime)s %(name)s %(levelname)-8s: %(message)s',
            datefmt='%m-%d %H:%M'))
        stream_handler.setLevel(level)
        logger.addHandler(stream_handler)

    if any([type(h) == logging.FileHandler for h in logger.handlers]) is False:
        if label:
            if output:
                if not os.path.exists(output):
                    os.makedirs(output, exist_ok=True)
            else:
                output = '.'
            log_file = '{}/{}.log'.format(output, label)
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(logging.Formatter(
                '%(asctime)s %(levelname)-8s: %(message)s', datefmt='%H:%M'))

            file_handler.setLevel(level)
            logger.addHandler(file_handler)

    for handler in logger.handlers:
        handler.setLevel(level)

    logger.info(f'Running Nessai version {version}')

    return logger


def is_jsonable(x):
    """
    Check if an object is JSON serialisable.

    Based on: https://stackoverflow.com/a/53112659

    Parameters
    ----------
    x : obj
        Object to check

    Returns
    -------
    bool
        Boolean that indicates if the object is JSON serialisable.
    """
    try:
        json.dumps(x)
        return True
    except (TypeError, OverflowError):
        return False


class FPJSONEncoder(json.JSONEncoder):
    """
    Class to encode numpy arrays and other non-serialisable objects.

    Based on: https://stackoverflow.com/a/57915246.

    Examples
    --------
    This class should be used in the ``cls`` argument::

        with open(filename, 'w') as wf:
             json.dump(d, wf, indent=4, cls=FPJSONEncoder)
    """

    def default(self, obj):
        """Method that returns a serialisable object"""
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif not is_jsonable(obj):
            return str(obj)
        else:
            return super(FPJSONEncoder, self).default(obj)


def safe_file_dump(data, filename, module, save_existing=False):
    """Safely dump data to a .pickle file

    See Bilby for the original impletmentation:
    https://git.ligo.org/michael.williams/bilby/-/blob/master/bilby/core/utils.py

    Parameters
    ----------
    data :
        Data to dump.
    filename : str
        The file to dump to.
    module : {pickle, dill}
        The python module to use.
    save_existing : bool, optional
        If true move the existing file to <file>.old.
    """
    if save_existing:
        if os.path.exists(filename):
            old_filename = filename + ".old"
            shutil.move(filename, old_filename)

    temp_filename = filename + ".temp"
    with open(temp_filename, "wb") as file:
        module.dump(data, file)
    shutil.move(temp_filename, filename)


def save_live_points(live_points, filename):
    """
    Save live points to a file using JSON.

    Live points are converted to a dictionary and then saved.

    Parameters
    ----------
    live_points : ndarray
        Live points to save.
    filename : str
        File to save to.
    """
    d = live_points_to_dict(live_points)
    with open(filename, 'w') as wf:
        json.dump(d, wf, indent=4, cls=FPJSONEncoder)


def configure_threads(max_threads=None, pytorch_threads=None, n_pool=None):
    """
    Configure the number of threads available. This is necessary when using
    PyTorch on the CPU as by default it will use all available threads.

    Notes
    -----
    Uses ``torch.set_num_threads``. If pytorch threads is None but other
    arguments are specified them the value is inferred from them.

    Parameters
    ----------
    max_threads: int, optional
        Maximum total number of threads to use between PyTorch and
        multiprocessing.
    pytorch_threads: int, optional
        Maximum number of threads for PyTorch on CPU.
    n_pool: int, optional
        Number of pools to use if using multiprocessing.
    """
    if max_threads is not None:
        if pytorch_threads is not None and pytorch_threads > max_threads:
            raise RuntimeError(
                f'More threads assigned to PyTorch ({pytorch_threads}) '
                f'than are available ({max_threads})')
        if n_pool is not None and n_pool >= max_threads:
            raise RuntimeError(
                f'More threads assigned to pool ({n_pool}) than are '
                f'available ({max_threads})')
        if (n_pool is not None and pytorch_threads is not None and
                (pytorch_threads + n_pool) > max_threads):
            raise RuntimeError(
                f'More threads assigned to PyTorch ({pytorch_threads}) '
                f'and pool ({n_pool})than are available ({max_threads})')

    if pytorch_threads is None:
        if max_threads is not None:
            if n_pool is not None:
                pytorch_threads = max_threads - n_pool
            else:
                pytorch_threads = max_threads

    if pytorch_threads is not None:
        logger.debug(
            f'Setting maximum number of PyTorch threads to {pytorch_threads}')
        torch.set_num_threads(pytorch_threads)


def _hist_bin_fd(x):
    """
    The Freedman-Diaconis histogram bin estimator.

    See original Numpy implementation.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.
    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    iqr = np.subtract(*np.percentile(x, [75, 25]))
    return 2.0 * iqr * x.size ** (-1.0 / 3.0)


def _hist_bin_sturges(x):
    """
    Sturges histogram bin estimator.

    See original Numpy implementation.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.
    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    return np.ptp(x) / (np.log2(x.size) + 1.0)


def auto_bins(x, max_bins=50):
    """
    Compute the number bins for a histogram using numpy.histogram_bin_edges
    but enforece a maximum number of bins.

    Parameters
    ----------
    array : array_like
        Input data
    bins : int or sequence of scalars or str, optional
        Method for determining number of bins, see numpy documentation
    max_bins : int, optional (1000)
        Maximum number of bins

    Returns
    -------
    int
        Number of bins
    """
    x = np.asarray(x)
    fd_bw = _hist_bin_fd(x)
    sturges_bw = _hist_bin_sturges(x)
    if fd_bw:
        bw = min(fd_bw, sturges_bw)
    else:
        bw = sturges_bw

    if bw:
        n_bins = int(np.ceil(np.ptp(x)) / bw)
    else:
        n_bins = 1

    nbins = min(n_bins, max_bins)
    assert isinstance(nbins, int)
    return nbins


def determine_rescaled_bounds(prior_min, prior_max, x_min, x_max, invert,
                              offset=0, rescale_bounds=[-1, 1],
                              inversion=False):
    """
    Determine the values of the prior min and max in the rescaled
    space.

    Parameters
    ----------
    prior_min : float
        Mininum of the prior.
    prior_max : float
        Maximum of the prior.
    x_min : float
        New minimum.
    x_max : float
        New maximum.
    invert : False or {'upper', 'lower', 'both'}
        Type of inversion.
    """
    if x_min == x_max:
        raise RuntimeError('New minimum and maximum are equal')
    scale = rescale_bounds[1] - rescale_bounds[0]
    shift = rescale_bounds[0]
    lower = scale * (prior_min - offset - x_min) / (x_max - x_min) + shift
    upper = scale * (prior_max - offset - x_min) / (x_max - x_min) + shift
    if not inversion:
        return lower, upper
    elif (not invert or invert is None):
        return 2 * lower - 1, 2 * upper - 1
    elif invert == 'upper':
        return lower - 1, 1 - lower
    elif invert == 'lower':
        return -upper, upper
    elif invert == 'both':
        return -0.5, 1.5
    else:
        raise RuntimeError



# -*- coding: utf-8 -*-
"""
Test the analytic proposal method.
"""
import pytest

from nessai.proposal import AnalyticProposal


@pytest.fixture(scope='function')
def proposal(model):
    return AnalyticProposal(model)


def test_populate(proposal):
    """Test the population process"""
    N = 500
    proposal.populate(N=N)

    assert proposal.samples.size == N
    assert proposal.populated


def test_draw(model, proposal):
    """Test the draw method"""
    N = 10
    old_point = model.new_point()
    assert not proposal.populated
    proposal.draw(old_point, N=N)
    assert proposal.populated
    [proposal.draw(old_point) for _ in range(N - 1)]
    assert not proposal.indices
    assert not proposal.populated



# -*- coding: utf-8 -*-
"""
Plotting utilities.
"""
import logging
from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd

from .utils import auto_bins

sns.set()
sns.set_style('ticks')

logger = logging.getLogger(__name__)

pairplot_kwargs = dict(corner=True, kind='scatter',
                       diag_kws=dict(histtype='step', bins='auto', lw=1.5,
                                     density=True, color='teal'),
                       plot_kws=dict(s=1.0, edgecolor=None, palette='viridis',
                                     color='teal'))


def plot_live_points(live_points, filename=None, bounds=None, c=None,
                     **kwargs):
    """
    Plot a set of live points in a corner-like plot.

    Parameters
    ----------
    live_points : ndarray
        Structured array of live points to plot.
    filename : str
        Filename for resulting figure
    bounds : dict:
        Dictionary of lower and upper bounds to plot
    c : str, optional
        Name of field in the structured array to use as the hue when plotting
        the samples. If not specified, no hue is used.
    kwargs :
        Keyword arguments used to update the pairplot kwargs. Diagonal and off-
        diagonal plots can be configured with ``diag_kws`` and ``plot_kws``.
    """
    pairplot_kwargs.update(kwargs)

    df = pd.DataFrame(live_points)
    df = df[np.isfinite(df).all(1)]

    if c is not None:
        hue = df[c]
    else:
        hue = None
    fig = sns.PairGrid(df, corner=True, diag_sharey=False)
    fig.map_diag(plt.hist, **pairplot_kwargs['diag_kws'])
    fig.map_offdiag(sns.scatterplot, hue=hue,
                    **pairplot_kwargs['plot_kws'])

    if bounds is not None:
        for i, v in enumerate(bounds.values()):
            fig.axes[i, i].axvline(v[0], ls=':', alpha=0.5, color='k')
            fig.axes[i, i].axvline(v[1], ls=':', alpha=0.5, color='k')

    if filename is not None:
        fig.savefig(filename)
        plt.close()
    else:
        return fig


def plot_1d_comparison(*live_points, parameters=None, labels=None,
                       bounds=None, hist_kwargs={},
                       filename=None, convert_to_live_points=False):
    """
    Plot 1d histograms comparing different sets of live points

    Parameters
    ----------
    *live_points : iterable of ndarrays
        Variable length argument of live points in structured arrays with
        fields. Also see ``parameters`` argument.
    parameters : array_like, optional
        Array of parameters (field names) to plot. Default None implies all
        fields are plotted.
    labels : array_like, optional
        Array of labels for each structured array being plotted (default None).
        If None each set of live points is labelled numerically
    bounds : dict, optional
        Dictionary of upper and lowers bounds to plot. Each key must
        match a field and each value must be an interable of length 2 in order
        lower then upper bound. If None (default), no bounds plotted.
    hist_kwargs : dict, optional
        Dictionary of keyword arguments passed to matplotlib.pyplot.hist.
    filename : str, optional
        Name of file for saving figure. (Default None implies figure is not
        saved).
    convert_to_live_points : bool, optional
        Set to true if inputs are not structured arrays of live points
    """
    if convert_to_live_points:
        live_points = list(live_points)
        if parameters is None:
            parameters = [i for i in range(live_points[0].shape[-1])]
        for i in range(len(live_points)):
            live_points[i] = \
                {k: v for k, v in zip(parameters, live_points[i].T)}

    elif any(lp.dtype.names is None for lp in live_points):
        raise RuntimeError('Live points are not structured arrays'
                           'Set `convert_to_live_points=True`.')
    elif parameters is None:
        parameters = live_points[0].dtype.names

    if labels is None:
        labels = [str(i) for i in range(len(live_points))]

    fig, axs = plt.subplots(len(parameters), 1, sharey=False,
                            figsize=(3, 3 * len(parameters)))

    axs = axs.ravel()
    for i, f in enumerate(parameters):
        xmin = np.min([np.min(lp[f][np.isfinite(lp[f])])
                       for lp in live_points])
        xmax = np.max([np.max(lp[f][np.isfinite(lp[f])])
                      for lp in live_points])
        for j, lp in enumerate(live_points):
            axs[i].hist(lp[f][np.isfinite(lp[f])],
                        bins=auto_bins(lp[f][np.isfinite(lp[f])]),
                        histtype='step',
                        range=(xmin, xmax), density=True, label=labels[j],
                        **hist_kwargs)
        axs[i].set_xlabel(f)
        if bounds is not None and f in bounds:
            axs[i].axvline(bounds[f][0], ls=':', alpha=0.5, color='k')
            axs[i].axvline(bounds[f][1], ls=':', alpha=0.5, color='k')

    if len(labels) > 1:
        handles, labels = plt.gca().get_legend_handles_labels()
        legend_labels = dict(zip(labels, handles))
        fig.legend(legend_labels.values(), legend_labels.keys(),
                   frameon=False, ncol=len(labels),
                   loc='upper center',
                   bbox_to_anchor=(0, 0.1 / len(parameters), 1, 1),
                   bbox_transform=plt.gcf().transFigure)

    plt.tight_layout()
    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')
        plt.close(fig)
    else:
        return fig


def plot_indices(indices, nlive=None, filename=None, plot_breakdown=True):
    """
    Histogram indices for index insertion tests, also includes the CDF.

    Parameters
    ----------
    indices : array_like
        List of insertion indices to plot
    nlive : int
        Number of live points used in the nested sampling run
    filename : str
        Filename used to save the figure.
    plot_breakdown : bool, optional
       If true, then the CDF for every nlive points is also plotted as grey
       lines.
    """
    indices = np.asarray(indices)
    if not indices.size or not nlive:
        logger.warning('Not producing indices plot.')
        return

    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    nbins = min(len(np.histogram_bin_edges(indices, 'auto')) - 1, 1000)
    if plot_breakdown:
        for i in range(len(indices) // nlive):
            ax[1].hist(indices[i * nlive:(i+1) * nlive], bins=nlive,
                       histtype='step', density=True, alpha=0.1, color='black',
                       lw=0.5, cumulative=True, range=(0, nlive-1))

    ax[0].hist(indices, density=True, color='tab:blue', linewidth=1.25,
               histtype='step', bins=nbins, label='produced',
               range=(0, nlive-1))
    ax[1].hist(indices, density=True, color='tab:blue', linewidth=1.25,
               histtype='step', bins=nlive, label='produced',
               cumulative=True, range=(0, nlive-1))

    if nlive is not None:
        ax[0].axhline(1 / nlive, color='black', linewidth=1.25,
                      linestyle='-', label='pmf', alpha=0.5)
        ax[0].axhline((1 + (nbins / len(indices)) ** 0.5) / nlive,
                      color='black', linewidth=1.25, linestyle=':', alpha=0.5,
                      label='1-sigma')
        ax[0].axhline((1 - (nbins / len(indices)) ** 0.5) / nlive,
                      color='black', linewidth=1.25, linestyle=':', alpha=0.5)
        ax[1].plot([0, nlive], [0, 1], color='black', linewidth=1.25,
                   linestyle=':', label='cmf')

    ax[0].legend(loc='lower right')
    ax[1].legend(loc='lower right')
    ax[0].set_xlim([0, nlive-1])
    ax[1].set_xlim([0, nlive-1])
    ax[0].set_xlabel('Insertion indices')
    ax[1].set_xlabel('Insertion indices')

    if filename is not None:
        plt.savefig(filename, bbox_inches='tight')
        plt.close(fig)
    else:
        return fig


def plot_loss(epoch, history, filename=None):
    """
    Plot the loss function per epoch.

    Parameters
    ----------
    epoch : int
        Final training epoch
    history : dict
        Dictionary with keys ``'loss'`` and ``'val_loss'``
    filename : str, optional
        Path for saving the figure. If not specified figure is returned
        instead.
    """
    fig, ax = plt.subplots()
    epochs = np.arange(1, epoch + 1, 1)
    plt.plot(epochs, history['loss'], label='Loss')
    plt.plot(epochs, history['val_loss'], label='Val. loss')
    plt.xlabel('Epochs')
    plt.ylabel('Negative log-likelihood')
    plt.legend()
    plt.tight_layout()
    if any(h < 0 for h in history['loss']):
        plt.yscale('symlog')
    else:
        plt.yscale('log')

    if filename is not None:
        plt.savefig(filename, bbox_inches='tight')
        plt.close(fig)
    else:
        return fig


def plot_trace(log_x, nested_samples, labels=None, filename=None):
    """
    Produce trace plot for all of the parameters.

    Parameters
    ----------
    log_x : array_like
        Array of log prior volumnes
    nested_samples : ndrray
        Array of nested samples to plot
    labels : list, optional
        List of labels to use instead of the names of parameters
    filename : str, optional
        Filename for saving the plot, if none plot is not saved and figure
        is returned instead.
    """
    nested_samples = np.asarray(nested_samples)
    names = nested_samples.dtype.names[:-2]

    if labels is None:
        labels = names

    if not len(labels) == len(names):
        raise RuntimeError(
            'Missing labels. List of labels does not have enough entries '
            f'({len(labels)}) for parameters: {nested_samples.dtype.names}')

    fig, axes = plt.subplots(len(labels), 1, figsize=(5, 3 * len(labels)),
                             sharex=True)
    axes = axes.ravel()

    for i, name in enumerate(names):
        axes[i].plot(log_x, nested_samples[name], ',')
        axes[i].set_ylabel(labels[i])

    axes[-1].set_xlabel('log X')
    axes[-1].invert_xaxis()

    if filename is not None:
        fig.savefig(filename, bbox_inches='tight')
        plt.close(fig)
    else:
        return fig



# -*- coding: utf-8 -*-
"""
Specific reparameterisations for gravitational-wave inference.
"""

import logging

from ..reparameterisations import (
    default_reparameterisations,
    Reparameterisation,
    RescaleToBounds,
    AnglePair
)

from ..priors import log_uniform_prior

from .utils import get_distance_converter


logger = logging.getLogger(__name__)


def get_gw_reparameterisation(reparameterisation):
    """
    Get a reparameterisation from the default list plus specific GW
    classes.
    """
    if isinstance(reparameterisation, str):
        rc, kwargs = default_gw.get(reparameterisation, (None, None))
        if rc is None:
            raise ValueError(
                f'Unknown GW reparameterisation: {reparameterisation}')
        else:
            if kwargs is None:
                kwargs = {}
            else:
                kwargs = kwargs.copy()
            return rc, kwargs
    elif (isinstance(reparameterisation, type) and
            issubclass(reparameterisation, Reparameterisation)):
        return reparameterisation, {}
    else:
        raise RuntimeError('Reparmeterisation must a str or class that '
                           'inherits from `Reparameterisation`')


class DistanceReparameterisation(RescaleToBounds):
    """Reparameterisation for distance.

    If the prior is specified and is one of the known priors then a rescaling
    is applied such that the resulting parameter has a uniform prior. If the
    prior is not specified, then the distance is rescaled an inversion is
    allowed on only the upper bound.

    Known priors
    ------------
    * Power-law: requires specifying the power. See converter kwargs.
    * Uniform-comoving-volume: uses a lookup table to convert to co-moving
    distance.

    Parameters
    ----------
    parameters : str
        Name of distance parameter to rescale.
    prior : {'power-law', 'uniform-comoving-volume'}, optional
        Prior used for the distance parameter
    prior_bounds : tuple
        Tuple of lower and upper bounds on the prior
    conveter_kwargs : dict, optional
        Keyword arguments parsed to converter object that converts the distance
        to a parameter with a uniform prior.
    allowed_bounds : list, optional
        List of the allowed bounds for inversion
    kwargs :
        Additional kwargs are parsed to the parent class.
    """
    def __init__(self, parameters=None, allowed_bounds=['upper'],
                 allow_both=False, converter_kwargs=None, prior=None,
                 prior_bounds=None, **kwargs):

        if isinstance(parameters, str):
            parameters = [parameters]

        if len(parameters) > 1:
            raise RuntimeError('Distance on supports one parameter')

        dc_class = get_distance_converter(prior)

        if converter_kwargs is None:
            converter_kwargs = {}
        self.distance_converter = dc_class(
            d_min=prior_bounds[parameters[0]][0],
            d_max=prior_bounds[parameters[0]][1],
            **converter_kwargs)

        pre_rescaling = (self.distance_converter.to_uniform_parameter,
                         self.distance_converter.from_uniform_parameter)

        super().__init__(parameters=parameters, prior=prior,
                         prior_bounds=prior_bounds,
                         pre_rescaling=pre_rescaling, **kwargs)

        if self.distance_converter.has_conversion:
            self._prime_prior = log_uniform_prior
            self.has_prime_prior = True
            if not self.distance_converter.has_jacobian:
                logger.debug('Distance converter does not have Jacobian, '
                             'require prime prior')
                self.requires_prime_prior = True
            self.update_prime_prior_bounds()
        else:
            self.has_prime_prior = False

        self.detect_edges_kwargs['allowed_bounds'] = allowed_bounds
        self.detect_edges_kwargs['allow_both'] = allow_both
        self.detect_edges_kwargs['x_range'] = \
            self.prior_bounds[self.parameters[0]]


default_gw = {
    'distance': (DistanceReparameterisation, {'boundary_inversion': True,
                                              'detect_edges': True,
                                              'inversion_type': 'duplicate'}),
    'time': (RescaleToBounds, {'offset': True, 'update_bounds': True}),
    'sky-ra-dec': (AnglePair, {'convention': 'ra-dec'}),
    'sky-az-zen': (AnglePair, {'convention': 'az-zen'}),
    'mass_ratio': (RescaleToBounds, {'detect_edges': True,
                                     'boundary_inversion': True,
                                     'inversion_type': 'duplicate',
                                     'update_bounds': True}),
    'mass': (RescaleToBounds, {'update_bounds': True}),
}


default_gw.update(default_reparameterisations)




import pytest
from scipy.stats import norm
import torch

from nessai.model import Model

cuda = pytest.mark.skipif(not torch.cuda.is_available(),
                          reason="test requires CUDA")


@pytest.fixture(scope='session')
def model():
    class TestModel(Model):

        def __init__(self):
            self.bounds = {'x': [-5, 5], 'y': [-5, 5]}
            self.names = ['x', 'y']

        def log_prior(self, x):
            log_p = 0.
            for n in self.names:
                log_p += ((x[n] >= self.bounds[n][0])
                          & (x[n] <= self.bounds[n][1])) \
                        / (self.bounds[n][1] - self.bounds[n][0])
            return log_p

        def log_likelihood(self, x):
            log_l = 0
            for pn in self.names:
                log_l += norm.logpdf(x[pn])
            return log_l

    return TestModel()


@pytest.fixture(scope='session')
def flow_config():
    d = dict(
            max_epochs=5,
            model_config=dict(n_blocks=2, n_neurons=2, n_layers=1,
                              device_tag='cpu',
                              kwargs=dict(batch_norm_between_layers=False))
            )
    return d


_requires_dependency_cache = dict()


def requires_dependency(name):
    """Decorator to declare required dependencies for tests.

    See git issue for original implementation:
    https://github.com/astropy/astropy/issues/5543

    Examples
    --------

    ::

        from gammapy.utils.testing import requires_dependency

        @requires_dependency('scipy')
        def test_using_scipy():
            import scipy
            ...
    """
    if name in _requires_dependency_cache:
        skip_it = _requires_dependency_cache[name]
    else:
        try:
            __import__(name)
            skip_it = False
        except ImportError:
            skip_it = True

        _requires_dependency_cache[name] = skip_it

    reason = 'Missing dependency: {}'.format(name)
    return pytest.mark.skipif(skip_it, reason=reason)



# -*- coding: utf-8 -*-
"""
Tests to ensure that the new GW proposal is equivalent to the old method
"""
import copy
import logging

import numpy as np

import pytest
from conftest import requires_dependency

from nessai.model import Model as BaseModel
from nessai.gw.proposal import GWFlowProposal, LegacyGWFlowProposal
from nessai.flowsampler import FlowSampler
from nessai.livepoint import dict_to_live_points

try:
    import bilby
except ImportError:
    print('Could not import bilby, some tests will not work')


logger = logging.getLogger(__name__)


@pytest.fixture(scope='module')
def injection_parameters():
    injection_parameters = dict(
        mass_1=36., mass_2=29., a_1=0.4, a_2=0.3, tilt_1=0.5, tilt_2=1.0,
        phi_12=1.7, phi_jl=0.3, luminosity_distance=2000., theta_jn=0.4,
        psi=2.659, phase=1.3, geocent_time=1126259642.413, ra=1.375,
        dec=-1.2108)
    return injection_parameters


@pytest.fixture(scope='function')
def legacy_kwargs():
    kwargs = {
        "boundary_inversion": True,
        "update_bounds": True,
        "rescale_parameters": True,
        "analytic_priors": True,
        "reparameterisations": {
            "rescale_angles": ["psi"],
            "inversion": True,
            "convert_to_angle": {"a_1": "split", "a_2": "duplicate"},
            "use_x_prime_prior": True,
            "uniform_distance_parameter": False,
            "spin_conversion": False
        },
        "detect_edges": True,
        "inversion_type": {"mass_ratio": "duplicate", "dc3": "duplicate"},
    }
    return kwargs


@pytest.fixture(scope='function')
def kwargs(legacy_kwargs):
    kwargs = copy.deepcopy(legacy_kwargs)
    kwargs['reparameterisations'] = {
        "mass_ratio": {"reparameterisation": "mass_ratio", "prior": "uniform"},
        "chirp_mass": {"reparameterisation": "mass", "prior": "uniform"},
        "luminosity_distance": {"reparameterisation": "distance",
                                "prior": "uniform-comoving-volume"},
        "geocent_time": {"reparameterisation": "time", "prior": "uniform",
                         "update_bounds": False},
        "sky-ra-dec": {"parameters": ["ra", "dec"], "prior": "isotropic"},
        "psi": {"reparameterisation": "angle-pi"},
        "theta_jn": {"reparameterisation": "angle-sine"},
        "tilt_1": {"reparameterisation": "angle-sine"},
        "tilt_2": {"reparameterisation": "angle-sine"},
        "phi_12": {"reparameterisation": "angle-2pi"},
        "phi_jl": {"reparameterisation": "angle-2pi"},
        "a_1": {"reparameterisation": "to-cartesian", "prior": "uniform"},
        "a_2": {"reparameterisation": "to-cartesian", "prior": "uniform",
                "mode": "duplicate"}
    }
    kwargs.pop('inversion_type')
    return kwargs


@requires_dependency('bilby')
@requires_dependency('lal')
@requires_dependency('astropy')
@pytest.mark.parametrize('parameters',
                         ['psi',
                          'theta_jn',
                          'tilt_1',
                          'tilt_2',
                          'phi_12',
                          'phi_jl',
                          'a_1',
                          'a_2',
                          ['ra', 'dec'],
                          ['chirp_mass', 'geocent_time'],
                          ['chirp_mass', 'mass_ratio'],
                          ['chirp_mass', 'luminosity_distance'],
                          ])
def test_parameter(parameters, injection_parameters, kwargs, legacy_kwargs,
                   tmpdir):
    """
    Test that the two proposal methods are equivalent for a set of parameters.

    Checks if the resulting x prime parameters are the same and their
    log-Jacobian determinants are also the same.
    """
    outdir = str(tmpdir.mkdir('test'))
    priors = bilby.gw.prior.BBHPriorDict()
    fixed_params = ["chirp_mass", "mass_ratio", "phi_12", "phi_jl", "a_1",
                    "a_2", "tilt_1", "tilt_2", "ra", "dec",
                    "luminosity_distance", "geocent_time", "theta_jn", "psi"]
    try:
        fixed_params.remove(parameters)
    except ValueError:
        for p in parameters:
            fixed_params.remove(p)
    priors['geocent_time'] = bilby.core.prior.Uniform(
            minimum=injection_parameters['geocent_time'] - 0.1,
            maximum=injection_parameters['geocent_time'] + 0.1,
            name='geocent_time', latex_label='$t_c$', unit='$s$')
    for key in fixed_params:
        if key in injection_parameters:
            priors[key] = injection_parameters[key]
        else:
            priors.pop(key)

    priors['mass_1'] = injection_parameters['mass_1']
    priors['mass_2'] = injection_parameters['mass_2']

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=1, sampling_frequency=256,
        frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole)

    likelihood = bilby.gw.GravitationalWaveTransient(
        interferometers=['H1'],
        waveform_generator=waveform_generator,
        priors=priors,
        phase_marginalization=True,
        distance_marginalization=False,
        time_marginalization=False, reference_frame='sky'
        )

    likelihood = bilby.core.likelihood.ZeroLikelihood(likelihood)

    def log_prior(theta):
        return priors.ln_prob(theta, axis=0)

    search_parameter_keys = []
    for key in priors:
        if isinstance(priors[key], bilby.core.prior.Prior) \
                and priors[key].is_fixed is False:
            search_parameter_keys.append(key)

    def log_likelihood(theta):
        params = {
            key: t for key, t in zip(search_parameter_keys, theta)}

        likelihood.parameters.update(params)
        return likelihood.log_likelihood_ratio()

    class Model(BaseModel):
        def __init__(self, names, priors):
            self.names = names
            self.priors = priors
            self._update_bounds()

        @staticmethod
        def log_likelihood(x, **kwargs):
            theta = [x[n].item() for n in search_parameter_keys]
            return log_likelihood(theta)

        @staticmethod
        def log_prior(x, names=None, **kwargs):
            if names is None:
                names = search_parameter_keys
            theta = {n: x[n] for n in names}
            return log_prior(theta)

        def _update_bounds(self):
            self.bounds = \
                {key: [self.priors[key].minimum, self.priors[key].maximum]
                 for key in self.names}

        def new_point(self, N=1):
            prior_samples = self.priors.sample(size=N)
            samples = {n: prior_samples[n] for n in self.names}
            self._update_bounds()
            return dict_to_live_points(samples)

        def new_point_log_prob(self, x):
            return self.log_prior(x)

    reparameterisations = kwargs.pop('reparameterisations')
    for k in list(reparameterisations.keys()):
        if k == 'sky-ra-dec':
            if any(p in fixed_params for p in ['ra', 'dec']):
                del reparameterisations['sky-ra-dec']
        elif k not in search_parameter_keys:
            del reparameterisations[k]

    if 'luminosity_distance' in parameters:
        legacy_kwargs['reparameterisations']['uniform_distance_parameter'] = \
                True

    model = Model(search_parameter_keys, priors)
    sampler = FlowSampler(model, output=outdir,
                          flow_class=LegacyGWFlowProposal, **legacy_kwargs)

    new_sampler = FlowSampler(model, output=outdir,
                              reparameterisations=reparameterisations,
                              flow_class=GWFlowProposal, **kwargs)

    sampler.ns._flow_proposal.initialise()
    new_sampler.ns._flow_proposal.initialise()

    sampler.ns._uninformed_proposal.populate(N=100)
    x = sampler.ns._uninformed_proposal.samples

    orig_proposal = sampler.ns._flow_proposal
    new_proposal = new_sampler.ns._flow_proposal

    tests = ['lower', 'upper', False]
    test_results = {t: False for t in tests}
    for test in tests:
        orig_proposal.check_state(x)
        np.random.seed(1234)
        x_prime, lj = orig_proposal.rescale(x, test=test)
        log_p = orig_proposal.x_prime_log_prior(x_prime)

        new_proposal.check_state(x)
        np.random.seed(1234)
        x_prime_new, lj_new = new_proposal.rescale(x, test=test)
        log_p_new = new_proposal.x_prime_log_prior(x_prime_new)

        np.testing.assert_array_equal(lj, lj_new)
        np.testing.assert_array_equal(log_p, log_p_new)
        try:
            np.testing.assert_array_equal(x_prime, x_prime_new)
        except AssertionError:
            flag = {n: False for n in x_prime.dtype.names}
            for n in x_prime.dtype.names:
                for nn in x_prime_new.dtype.names:
                    if np.allclose(x_prime[n], x_prime_new[nn]):
                        logger.critical(f'{n} is equivalent to {nn}')
                        flag[n] = True

            assert all(v for v in flag.values()), print(f'Flags: {flag}')
        test_results[test] = True

    assert all(t for t in test_results.values())

    if getattr(orig_proposal, '_rescaled_min'):
        logger.info('Checking prime prior bounds')
        test_results = {t: False for t in tests}
        for test in tests:
            logger.info(f'Testing with inversion and test={test}')
            orig_proposal.check_state(x)
            np.random.seed(1234)
            t, lj = orig_proposal.rescale(x, test=test)
            new_proposal.check_state(x)
            np.random.seed(1234)
            t_new, lj_new = new_proposal.rescale(x, test=test)
            min_vals = \
                {n: False for n in new_proposal._reparameterisation.keys()}
            max_vals = \
                {n: False for n in new_proposal._reparameterisation.keys()}
            for k, r in new_proposal._reparameterisation.items():
                for pb in r.prime_prior_bounds.values():
                    for vmin in orig_proposal._rescaled_min.values():
                        logger.warning([k, pb, vmin])
                        if np.isclose(pb[0], vmin):
                            min_vals[k] = True
                    for vmax in orig_proposal._rescaled_max.values():
                        logger.warning([k, pb, vmax])
                        if np.isclose(pb[1], vmax):
                            max_vals[k] = True

            assert all(t for t in min_vals.values()), print(min_vals)
            assert all(t for t in max_vals.values()), print(max_vals)
            test_results[test] = True

        assert all(t for t in test_results.values())



# -*- coding: utf-8 -*-
"""
Test the rejection proposal class.
"""
import pytest

from nessai.proposal import RejectionProposal


@pytest.fixture()
def poolsize():
    return 10


@pytest.fixture(scope='function')
def proposal(model, poolsize):
    return RejectionProposal(model, poolsize=poolsize)


def test_populate(proposal, poolsize):
    """Test the population process"""
    proposal.populate(N=poolsize)

    assert proposal.samples.size == poolsize
    assert proposal.populated


def test_draw(model, proposal, poolsize):
    """Test the draw method"""
    old_point = model.new_point()
    assert not proposal.populated
    proposal.draw(old_point)
    assert proposal.populated
    [proposal.draw(old_point) for _ in range(poolsize - 1)]
    assert not proposal.indices
    assert not proposal.populated



# -*- coding: utf-8 -*-
"""
Functions realted to computing the evidence and posterior samples.
"""
import numpy as np


def logsubexp(x, y):
    """
    Helper function to compute the exponential
    of a difference between two numbers

    Computes: ``x + np.log1p(-np.exp(y-x))``

    Parameters
    ----------
    x, y : float or array_like
        Inputs
    """
    if np.any(x < y):
        raise RuntimeError('cannot take log of negative number '
                           f'{str(x)!s} - {str(y)!s}')

    return x + np.log1p(-np.exp(y - x))


def log_integrate_log_trap(log_func, log_support):
    """
    Trapezoidal integration of given log(func). Returns log of the integral.

    Parameters
    ----------
    log_func : array_like
        Log values of the function to integrate over.
    log_support : array_like
        Log-evidences for each value.

    Returns
    -------
    float
        Log of the result of the integral.
    """
    log_func_sum = np.logaddexp(log_func[:-1], log_func[1:]) - np.log(2)
    log_dxs = logsubexp(log_support[:-1], log_support[1:])

    return np.logaddexp.reduce(log_func_sum + log_dxs)


def compute_weights(samples, nlive):
    """
    Returns the log-evidence and log-weights for the log-likelihood samples
    assumed to the result of nested sampling with nlive live points

    Parameters
    ----------
    samples : array_like
        Log-likelihood samples.
    nlive : int
        Number of live points used in nested sampling.

    Returns
    -------
    float
        The computed log-evidence.
    array_like
        Array of computed weigths (already normalised by the log-evidence).
    """
    samples = np.asarray(samples)
    start_data = np.concatenate(([float('-inf')], samples[:-nlive]))
    end_data = samples[-nlive:]

    log_wts = np.zeros(samples.shape[0])

    log_vols_start = np.cumsum(np.ones(len(start_data) + 1)
                               * np.log1p(-1. / nlive)) - np.log1p(-1 / nlive)
    log_vols_end = np.zeros(len(end_data))
    log_vols_end[-1] = np.NINF
    log_vols_end[0] = log_vols_start[-1] + np.log1p(-1.0 / nlive)
    for i in range(len(end_data) - 1):
        log_vols_end[i+1] = log_vols_end[i] + np.log1p(-1.0 / (nlive - i))

    log_likes = np.concatenate((start_data, end_data, [end_data[-1]]))

    log_vols = np.concatenate((log_vols_start, log_vols_end))
    log_ev = log_integrate_log_trap(log_likes, log_vols)

    log_dXs = logsubexp(log_vols[:-1], log_vols[1:])
    log_wts = log_likes[1:-1] + log_dXs[:-1]

    log_wts -= log_ev

    return log_ev, log_wts


def draw_posterior_samples(nested_samples, nlive):
    """
    Draw posterior samples given the nested samples and number of live points.

    Parameters
    ----------
    nested_samples : structured array
        Array of nested samples.
    nlive : int
        Number of live points used during nested sampling.

    Returns
    -------
    array_like
        Samples from the posterior distribution.
    """
    log_Z, log_w = compute_weights(nested_samples['logL'], nlive)
    log_w -= np.max(log_w)
    log_u = np.log(np.random.rand(nested_samples.size))
    return nested_samples[log_w > log_u]




import numpy as np
from scipy import stats
import torch
import pytest

from nessai.flows.distributions import MultivariateNormal


@pytest.fixture
def dims():
    """Number of dimensions"""
    return 4


@pytest.fixture
def var():
    """Variance to test"""
    return 2


@pytest.fixture
def dist(dims, var):
    """Instance of MultivariateNormal"""
    return MultivariateNormal([dims], var=var)


@pytest.fixture
def scipy_dist(dims, var):
    mean = np.zeros(dims)
    var = var * np.eye(dims)
    return stats.multivariate_normal(mean=mean, cov=var)


def test_log_prob(dist, scipy_dist, dims):
    """Test the log probability of the multivariate normal"""
    x = np.random.rand(1000, dims)
    log_prob = dist._log_prob(torch.from_numpy(x).float(), None).numpy()
    np.testing.assert_array_almost_equal(log_prob, scipy_dist.logpdf(x))


def test_log_prob_invalid_shape(dist, dims):
    """
    Assert the the log prob raises an error for inputs that are the incorrect
    shape.
    """
    x = np.random.rand(10, dims + 2)
    with pytest.raises(ValueError) as excinfo:
        dist._log_prob(torch.from_numpy(x).float(), None)

    assert 'Expected input of shape' in str(excinfo.value)


@pytest.mark.flaky(run=5)
def test_sample(dist, scipy_dist):
    """
    Test the sample method and check if the resulting samples pass a
    KS test using the scipy distribution
    """
    samples = dist._sample(1000, None).numpy()
    p, _ = stats.kstest(samples, scipy_dist.cdf)
    assert p > 0.05


def test_sample_context(dist):
    """
    Test the sample method with a context that is not None.
    Should raise NotImplementedError
    """
    with pytest.raises(NotImplementedError):
        dist._sample(1000, True)


def test_mean(dist, dims):
    """Test the mean of the multivariate normal"""
    np.testing.assert_array_equal(dist._mean(None).numpy(), np.zeros(dims))


def test_mean_context(dist):
    """
    Test the mean of the multivariate normal with context raises
    a not implemneted error
    """
    with pytest.raises(NotImplementedError):
        dist._mean(True)



# -*- coding: utf-8 -*-
"""
Code related to the implementation of normalising flows and some complete
implementations.
"""
from .base import BaseFlow, NFlow
from .maf import MaskedAutoregressiveFlow
from .nsf import NeuralSplineFlow
from .realnvp import FlexibleRealNVP
from .utils import setup_model, reset_weights, reset_permutations

__all__ = ["BaseFlow",
           "NFlow",
           "MaskedAutoregressiveFlow",
           "NeuralSplineFlow",
           "FlexibleRealNVP",
           "setup_model",
           "reset_weights",
           "reset_permutations"]




import numpy as np
from scipy import stats
import pytest

import nessai.utils as utils

from conftest import cuda


@pytest.mark.parametrize("x, y, log_J", [(0., -np.inf, np.inf),
                                         (1., np.inf, np.inf)])
def test_logit_bounds(x, y, log_J):
    """
    Test logit at the bounds
    """
    with pytest.warns(RuntimeWarning):
        assert utils.logit(x, fuzz=0) == (y, log_J)


@pytest.mark.parametrize("x, y, log_J", [(np.inf, 1, -np.inf),
                                         (-np.inf, 0, -np.inf)])
def test_sigmoid_bounds(x, y, log_J):
    """
    Test sigmoid for inf
    """
    assert utils.sigmoid(x, fuzz=0) == (y, log_J)


@pytest.mark.parametrize("p", [1e-5, 0.5, 1 - 1e-5])
def test_logit_sigmoid(p):
    """
    Test invertibility of sigmoid(logit(x))
    """
    x = utils.logit(p, fuzz=0)
    y = utils.sigmoid(x[0], fuzz=0)
    np.testing.assert_equal(p, y[0])
    np.testing.assert_almost_equal(x[1] + y[1], 0)


@pytest.mark.parametrize("p", [-10, -1, 0, 1, 10])
def test_sigmoid_logit(p):
    """
    Test invertibility of logit(sigmoid(x))
    """
    x = utils.sigmoid(p, fuzz=0)
    y = utils.logit(x[0], fuzz=0)
    np.testing.assert_almost_equal(p, y[0])
    np.testing.assert_almost_equal(x[1] + y[1], 0)


def test_replace_in_list():
    """
    Test if the list produced contains the correct entries in the correct
    locations
    """
    x = [1, 2, 3]
    utils.replace_in_list(x, [1, 2], [5, 4])
    assert x == [5, 4, 3]


def test_replace_in_list_item():
    """
    Test if items are correctly converted to lists in replace_in_list function
    """
    x = [1, 2, 3]
    utils.replace_in_list(x, 3, 4)
    assert x == [1, 2, 4]


@pytest.mark.parametrize("mode", ['D+', 'D-'])
def test_ks_test(mode):
    """
    Test KS test for insertion indices with a specifed mode
    """
    indices = np.random.randint(0, 1000, 1000)
    out = utils.compute_indices_ks_test(indices, 1000, mode=mode)
    assert all([o > 0. for o in out])


def test_ks_test_undefined_mode():
    """
    Test KS test for insertion indices with undefined mode
    """
    indices = np.random.randint(0, 1000, 1000)
    with pytest.raises(RuntimeError):
        utils.compute_indices_ks_test(indices, 1000, mode='two-sided')


def test_ks_test_empty_indices():
    """
    Test KS test for insertion indices with empty input array
    """
    out = utils.compute_indices_ks_test([], 1000, mode='D+')
    assert all(o is None for o in out)


@pytest.mark.parametrize("ndims, radius", [(2, 1), (10, 2), (10, 10), (1, 1)])
def test_draw_surface_nsphere(ndims, radius):
    """
    Assert the correct number of samples with the correct dimensions are drawn
    and all lie on the surface.
    """
    out = utils.draw_surface_nsphere(ndims, r=radius, N=1000)

    assert out.shape[0] == 1000
    assert out.shape[1] == ndims
    np.testing.assert_allclose(np.sqrt(np.sum(out ** 2., axis=1)), radius)


@pytest.mark.parametrize("ndims, radius", [(2, 1), (10, 2), (10, 10), (1, 1)])
def test_draw_nball(ndims, radius):
    """
    Assert the correct number of samples with the correct dimensions are drawn
    and all lie within the n-ball.
    """
    out = utils.draw_nsphere(ndims, r=radius, N=1000)

    assert out.shape[0] == 1000
    assert out.shape[1] == ndims
    np.testing.assert_array_less(np.sqrt(np.sum(out ** 2, axis=-1)), radius)


def test_get_uniform_distribution_cpu():
    """
    Test function for getting uniform torch distrbution over n dimensions
    when called on cpu
    """
    dist = utils.get_uniform_distribution(10, 1, 'cpu')
    assert dist.sample().get_device() == -1


@cuda
def test_get_uniform_distribution_cuda():
    """
    Test function for getting uniform torch distrbution over n dimensions
    when called on CUDA
    """
    dist = utils.get_uniform_distribution(10, 1, device='cuda')
    assert dist.sample().get_device() != -1


@pytest.mark.parametrize("r, var, fuzz", [
    (1.0, 1.0, 1.0), (2.0, 1.0, 1.0), (2.0, 2.0, 1.0), (4.0, 2.0, 1.5),
    (7.0, 4.0, 2.0)])
@pytest.mark.flaky(run=5)
def test_draw_truncated_gaussian_1d(r, var, fuzz):
    """
    Test drawing from a truncated Gaussian in 1d
    """
    s = utils.draw_truncated_gaussian(1, r, var=var, N=1000, fuzz=fuzz)
    sigma = np.sqrt(var)
    d = stats.truncnorm(-r * fuzz / sigma, r * fuzz / sigma,
                        loc=0, scale=sigma)
    _, p = stats.kstest(np.squeeze(s), d.cdf)
    assert p >= 0.05



# -*- coding: utf-8 -*-
"""
Distributions to use as the 'base distribution' for normalising flows.
"""

from nflows.distributions import Distribution
from nflows.utils import torchutils
import numpy as np
import torch


class MultivariateNormal(Distribution):
    """
    A multivariate Normal with zero mean and specified covariance.

    Parameters
    ----------
    shape : tuple
        Shape of distribution, this is used to determine the number of
        dimensions.
    var : float, optional
        Variance of the distrinution.
    """

    def __init__(self, shape, var=1):
        super().__init__()
        self._shape = torch.Size(shape)
        self._var = var

        self.register_buffer(
            "_log_z",
            torch.tensor(0.5 * np.prod(shape) * np.log(2 * np.pi * var),
                         dtype=torch.float64),
            persistent=False)

    def _log_prob(self, inputs, context):
        # Note: the context is ignored.
        if inputs.shape[1:] != self._shape:
            raise ValueError(
                "Expected input of shape {}, got {}".format(
                    self._shape, inputs.shape[1:]
                )
            )
        neg_energy = -(0.5 / self._var) * \
            torchutils.sum_except_batch(inputs ** 2, num_batch_dims=1)
        return neg_energy - self._log_z

    def _sample(self, num_samples, context):
        if context is None:
            return torch.normal(0, self._var,
                                size=(num_samples, *self._shape),
                                device=self._log_z.device)
        else:
            raise NotImplementedError

    def _mean(self, context):
        if context is None:
            return self._log_z.new_zeros(self._shape)
        else:
            raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Main code that handles running and checkpoiting the sampler.
"""
import json
import logging
import os
import signal
import sys

import numpy as np

from . import __version__ as version
from .livepoint import live_points_to_dict
from .nestedsampler import NestedSampler
from .posterior import draw_posterior_samples
from .utils import FPJSONEncoder, configure_threads


logger = logging.getLogger(__name__)


class FlowSampler:
    """
    Main class to handle running the nested sampler.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User-defined model.
    output : str, optional
        Output directory
    resume : bool, optional
        If True try to resume the sampler is the resume file existis.
    resume_file : str, optional
        File to resume sampler from.
    weights_files : str, optional
        Weights files used to resume sampler that replaces the weights file
        saved internally.
    max_threads : int, optional
        Maximum number of threads to use. If ``None`` torch uses all available
        threads.
    exit_code : int, optional
        Exit code to use when forceably exiting the sampler.
    kwargs :
        Keyword arguments passed to :obj:`~nessai.nestedsampler.NestedSampler`.
    """
    def __init__(self, model, output='./', resume=True,
                 resume_file='nested_sampler_resume.pkl', weights_file=None,
                 exit_code=130, max_threads=1, **kwargs):

        configure_threads(
            max_threads=max_threads,
            pytorch_threads=kwargs.get('pytorch_threads', None),
            n_pool=kwargs.get('n_pool', None)
            )

        self.exit_code = exit_code

        self.output = os.path.join(output, '')
        if resume:
            if not any((os.path.exists(os.path.join(self.output, f)) for f in
                        [resume_file, resume_file + '.old'])):
                logger.warning('No files to resume from, starting sampling')
                self.ns = NestedSampler(model, output=self.output,
                                        resume_file=resume_file, **kwargs)
            else:
                try:
                    self.ns = NestedSampler.resume(
                        os.path.join(self.output, resume_file), model,
                        kwargs['flow_config'], weights_file)
                except (FileNotFoundError, RuntimeError) as e:
                    logger.error(
                        f'Could not load resume file from: {self.output} '
                        f'with error {e}')
                    try:
                        resume_file += '.old'
                        self.ns = NestedSampler.resume(
                            os.path.join(self.output, resume_file), model,
                            kwargs['flow_config'], weights_file)
                    except RuntimeError as e:
                        logger.error('Could not load old resume '
                                     f'file from: {self.output}')
                        raise RuntimeError('Could not resume sampler '
                                           f'with error: {e}')
        else:
            self.ns = NestedSampler(model, output=self.output,
                                    resume_file=resume_file, **kwargs)

        self.save_kwargs(kwargs)

        try:
            signal.signal(signal.SIGTERM, self.safe_exit)
            signal.signal(signal.SIGINT, self.safe_exit)
            signal.signal(signal.SIGALRM, self.safe_exit)
        except AttributeError:
            logger.critical('Can not set signal attributes on this system')

    def run(self, plot=True, save=True):
        """
        Run the nested samper

        Parameters
        ----------
        plot : bool, optional
            Toggle plots produced once the sampler has converged
        save : bool, opitional
            Toggle automatic saving of results
        """
        self.ns.initialise()
        self.logZ, self.nested_samples = \
            self.ns.nested_sampling_loop()
        logger.info((f'Total sampling time: {self.ns.sampling_time}'))

        logger.info('Starting post processing')
        logger.info('Computing posterior samples')
        self.posterior_samples = draw_posterior_samples(self.nested_samples,
                                                        self.ns.nlive)
        logger.info(f'Returned {self.posterior_samples.size} '
                    'posterior samples')

        if save:
            self.save_results(f'{self.output}/result.json')

        if plot:
            from nessai import plot

            plot.plot_live_points(self.posterior_samples,
                                  filename=(f'{self.output}/'
                                            'posterior_distribution.png'))

            plot.plot_indices(self.ns.insertion_indices, self.ns.nlive,
                              filename=f'{self.output}/insertion_indices.png')

            self.ns.state.plot(f'{self.output}/logXlogL.png')

    def save_kwargs(self, kwargs):
        """
        Save the dictionary of keyword arguments used.

        Uses an encoder class to handle numpy arrays.

        Parameters
        ----------
        kwargs : dict
            Dictionary of kwargs to save.
        """
        d = kwargs.copy()
        with open(f'{self.output}/config.json', 'w') as wf:
            try:
                json.dump(d, wf, indent=4, cls=FPJSONEncoder)
            except TypeError:
                if 'flow_class' in d:
                    d['flow_class'] = str(d['flow_class'])
                    json.dump(d, wf, indent=4, cls=FPJSONEncoder)
            except Exception as e:
                raise e

    def save_results(self, filename):
        """
        Save the results from sampling to a specific JSON file.

        Parameters
        ----------
        filename : str
            Name of file to save results to.
        """
        iterations = np.arange(len(self.ns.min_likelihood)) \
            * (self.ns.nlive // 10)
        iterations[-1] = self.ns.iteration
        d = dict()
        d['version'] = version
        d['history'] = dict(
                iterations=iterations,
                min_likelihood=self.ns.min_likelihood,
                max_likelihood=self.ns.max_likelihood,
                likelihood_evaluations=self.ns.likelihood_evaluations,
                logZ=self.ns.logZ_history,
                dZ=self.ns.dZ_history,
                mean_acceptance=self.ns.mean_acceptance_history,
                rolling_p=self.ns.rolling_p,
                population=dict(
                    iterations=self.ns.population_iterations,
                    acceptance=self.ns.population_acceptance
                    ),
                training_iterations=self.ns.training_iterations

                )
        d['insertion_indices'] = self.ns.insertion_indices
        d['nested_samples'] = live_points_to_dict(self.nested_samples)
        d['posterior_samples'] = live_points_to_dict(self.posterior_samples)
        d['log_evidence'] = self.ns.log_evidence
        d['information'] = self.ns.information
        d['sampling_time'] = self.ns.sampling_time.total_seconds()
        d['training_time'] = self.ns.training_time.total_seconds()
        d['population_time'] = self.ns.proposal_population_time.total_seconds()
        if self.ns.likelihood_evaluation_time.total_seconds():
            d['likelihood_evaluation_time'] = \
                self.ns.likelihood_evaluation_time.total_seconds()

        with open(filename, 'w') as wf:
            json.dump(d, wf, indent=4, cls=FPJSONEncoder)

    def safe_exit(self, signum=None, frame=None):
        """
        Safely exit. This includes closing the multiprocessing pool.
        """
        logger.warning(f'Trying to safely exit with code {signum}')
        self.ns.proposal.close_pool(code=signum)
        self.ns.checkpoint()

        logger.warning(f'Exiting with code: {self.exit_code}')
        sys.exit(self.exit_code)



# -*- coding: utf-8 -*-
"""
Test the multiprocessing pool for the different proposal methods.
"""
from numpy.testing import assert_array_equal
import pytest

from nessai.proposal import AnalyticProposal, RejectionProposal, FlowProposal


@pytest.mark.parametrize('proposal', [AnalyticProposal, RejectionProposal,
                                      FlowProposal])
@pytest.mark.parametrize('n_pool', [None, 1])
def test_analytic_pool(proposal, n_pool, model, tmpdir):
    try:
        prop = proposal(model, n_pool=n_pool, poolsize=10, plot=False,
                        max_radius=2, output=str(tmpdir.mkdir('test')))
    except TypeError:
        prop = proposal(model, n_pool=n_pool, poolsize=10)

    try:
        prop.initialise()
    except AttributeError:
        pass

    prop.configure_pool()
    if n_pool:
        assert prop.pool
    else:
        assert prop.pool is None

    worst = model.new_point()
    prop.draw(worst)

    if n_pool:
        logL = [model.log_likelihood(s) for s in prop.samples]
        assert_array_equal(prop.samples['logL'], logL)
    else:
        assert all(ll == 0. for ll in prop.samples['logL'])



import numpy as np
import pytest
import torch

from nessai.proposal import FlowProposal
from nessai.livepoint import numpy_array_to_live_points

torch.set_num_threads(1)


@pytest.mark.parametrize('latent_prior', ['gaussian', 'truncated_gaussian',
                                          'uniform_nball', 'uniform_nsphere',
                                          'uniform'])
@pytest.mark.parametrize('expansion_fraction', [0, 1, None])
@pytest.mark.parametrize('check_acceptance', [False, True])
@pytest.mark.parametrize('rescale_parameters', [False, True])
@pytest.mark.parametrize('max_radius', [False, 2])
@pytest.mark.timeout(10)
@pytest.mark.flaky(run=3)
def test_flowproposal_populate(tmpdir, model, latent_prior, expansion_fraction,
                               check_acceptance, rescale_parameters,
                               max_radius):
    """
    Test the populate method in the FlowProposal class with a range of
    parameters
    """
    output = str(tmpdir.mkdir('flowproposal'))
    fp = FlowProposal(
        model,
        output=output,
        plot=False,
        poolsize=1000,
        latent_prior=latent_prior,
        expansion_fraction=expansion_fraction,
        check_acceptance=check_acceptance,
        rescale_parameters=rescale_parameters,
        max_radius=max_radius
    )

    fp.initialise()
    worst = numpy_array_to_live_points(0.5 * np.ones(fp.dims), fp.names)
    fp.populate(worst, N=100)

    assert fp.x.size == 100



# -*- coding: utf-8 -*-
"""
Functions and objects related to reparametersiations for use in the
``reparameterisations`` dictionary.

See the documentation for an in-depth description of how to use these
functions and classes.
"""
import logging

import numpy as np
from scipy import stats

from .utils import (
    detect_edge,
    configure_edge_detection,
    rescale_zero_to_one,
    inverse_rescale_zero_to_one,
    rescale_minus_one_to_one,
    inverse_rescale_minus_one_to_one,
    determine_rescaled_bounds,
    rescaling_functions
)

from .priors import (
    log_uniform_prior,
    log_2d_cartesian_prior,
    log_2d_cartesian_prior_sine,
    )


logger = logging.getLogger(__name__)


def get_reparameterisation(reparameterisation):
    """Function to get a reparmeterisation class from a name

    Parameters
    ----------
    reparameterisation : str, \
            :obj:`nessai.reparameterisations.Reparametersiation`
        Name of the reparameterisations to return or a class that inherits from
        :obj:`~nessai.reparameterisations.Reparmeterisation`

    Returns
    -------
    :obj:`nessai.reparameteristaions.Reparmeterisation`
        T
    dict
        Keyword arguments for the specific reparameterisations.
    """
    if isinstance(reparameterisation, str):
        rc, kwargs = default_reparameterisations.get(
            reparameterisation, (None, None))
        if rc is None:
            raise ValueError(
                f'Unknown reparameterisation: {reparameterisation}')
        else:
            if kwargs is None:
                kwargs = {}
            else:
                kwargs = kwargs.copy()
            return rc, kwargs

    elif (isinstance(reparameterisation, type) and
            issubclass(reparameterisation, Reparameterisation)):
        return reparameterisation, {}
    else:
        raise RuntimeError('Reparmeterisation must a str or class that '
                           'inherits from `Reparameterisation`')


class Reparameterisation:
    """
    Base object for reparameterisations.

    Parameters
    ----------
    parameters : str or list
        Name of parameters to reparameterise.
    """
    _update_bounds = False
    has_prior = False
    has_prime_prior = False
    requires_prime_prior = False

    def __init__(self, parameters=None, prior_bounds=None):
        if not isinstance(parameters, (str, list)):
            raise TypeError('Parameters must be a str or list.')

        self.parameters = \
            [parameters] if isinstance(parameters, str) else parameters

        if isinstance(prior_bounds, (list, tuple)):
            if len(prior_bounds) == 2:
                prior_bounds = {self.parameters[0]: np.asarray(prior_bounds)}
            else:
                raise RuntimeError('Prior bounds got a list of len > 2')
        elif not isinstance(prior_bounds, dict):
            raise TypeError('Prior bounds must be dict or tuple of len 2')

        if set(self.parameters) - set(prior_bounds.keys()):
            raise RuntimeError(
                'Mismatch between parameters and prior bounds: '
                f'{set(self.parameters)}, {set(prior_bounds.keys())}')

        self.prior_bounds = {p: np.asarray(b) for p, b in prior_bounds.items()}
        self.prime_parameters = [p + '_prime' for p in self.parameters]
        self.requires = []
        logger.debug(f'Initialised reparameterisation: {self.name}')

    @property
    def name(self):
        """Unique name of the reparameterisations"""
        return self.__class__.__name__.lower() + '_' + \
            '_'.join(self.parameters)

    def reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space to x'-space.

        Parameters
        ----------
        x : structured array
            Array of inputs
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError

    def inverse_reparameterise(self, x, x_prime, log_j):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated

        Returns
        -------
        x, x_prime : structured arrays
            Update version of the x and x_prime arrays
        log_j : array_like
            Updated log Jacobian determinant
        """
        raise NotImplementedError


class CombinedReparameterisation(dict):
    """Class to handle mulitple reparameterisations

    Parameters
    ----------
    reparameterisations : list, optional
        List of reparameterisations to add to the combined reparameterisations.
        Further reparameterisations can be added using
        `add_reparameterisations`.
    """
    def __init__(self, reparameterisations=None):
        super().__init__()
        self.reparameterisations = {}
        self.parameters = []
        self.prime_parameters = []
        self.requires = []
        if reparameterisations is not None:
            self.add_reparameterisations(reparameterisations)

    @property
    def has_prime_prior(self):
        """Boolean to check if prime prior can be enabled"""
        return all(r.has_prime_prior for r in self.values())

    @property
    def requires_prime_prior(self):
        """Boolean to check if any of the priors require the prime space"""
        return any(r.requires_prime_prior for r in self.values())

    def _add_reparameterisation(self, reparameterisation):
        requires = reparameterisation.requires
        if (requires and
                (requires not in self.parameters or
                    requires not in self.prime_parameters)):
            raise RuntimeError(
                f'Could not add {reparameterisation}, missing requirement(s): '
                f'{reparameterisation.requires}.')

        self[reparameterisation.name] = reparameterisation
        self.parameters += reparameterisation.parameters
        self.prime_parameters += reparameterisation.prime_parameters
        self.requires += reparameterisation.requires

    def add_reparameterisation(self, reparameterisation):
        """Add a reparameterisation"""
        self.add_reparameterisations(reparameterisation)

    def add_reparameterisations(self, reparameterisations):
        """Add multiple reparameterisations

        Parameters
        ----------
        reparameterisations : list of :`obj`:Reparameterisation
            List of reparameterisations to add.
        """
        if not isinstance(reparameterisations, list):
            reparameterisations = [reparameterisations]
        for r in reparameterisations:
            self._add_reparameterisation(r)

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for r in self.values():
            x, x_prime, log_j = r.reparameterise(x, x_prime, log_j, **kwargs)
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for r in reversed(list(self.values())):
            x, x_prime, log_j = r.inverse_reparameterise(
                x, x_prime, log_j, **kwargs)
        return x, x_prime, log_j

    def update_bounds(self, x):
        """
        Update the bounds used for the reparameterisation
        """
        for r in self.values():
            if hasattr(r, 'update_bounds'):
                logger.debug(f'Updating bounds for: {r.name}')
                r.update_bounds(x)

    def reset_inversion(self):
        """
        Reset edges for boundary inversion
        """
        for r in self.values():
            if hasattr(r, 'reset_inversion'):
                r.reset_inversion()

    def log_prior(self, x):
        """
        Compute any additional priors for auxiliary parameters
        """
        log_p = 0
        for r in self.values():
            if r.has_prior:
                log_p += r.log_prior(x)
        return log_p

    def x_prime_log_prior(self, x_prime):
        """
        Compute the prior in the prime space
        """
        log_p = 0
        for r in self.values():
            log_p += r.x_prime_log_prior(x_prime)
        return log_p


class NullReparameterisation(Reparameterisation):
    """Reparameteristion that does not modify the parameters"""
    def __init__(self, parameters=None):
        if not isinstance(parameters, (str, list)):
            raise TypeError('Parameters must be a str or list.')

        self.parameters = \
            [parameters] if isinstance(parameters, str) else parameters

        self.prime_parameters = self.parameters
        self.requires = []
        logger.debug(f'Initialised reparameterisation: {self.name}')

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x_prime[self.prime_parameters] = x[self.parameters]
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : Log jacobian to be updated
        """
        x[self.parameters] = x_prime[self.prime_parameters]
        return x, x_prime, log_j


class RescaleToBounds(Reparameterisation):
    """Reparmeterisation that maps to the specified interval.

    By default the interval is [-1, 1]. Also includes options for
    boundary inversion.

    This reparameterisation can handle multiple parameters.

    Parameters
    ----------
    parameters : list of str
        List of the names of parameters
    prior_bounds : dict
        Dictionary of prior bounds for each parameter. Does not need to be
        specified by the user.
    rescale_bounds : list of tuples, optional
        Bounds to rescale to.
    update_bounds : bool, optional
        Enable or disable updating bounds.
    prior : {'uniform', None}
        Type of prior used, if uniform prime prior is enabled.
    boundary_inversion : bool, list, dict, optional
        Configuration for boundary inversion. If a list, inversion is only
        applied to the parameters in the list based on `inversion_type`. If
        a `dict` then each item should be a parameter and a corresponding
        inversion type `{'split', 'inversion'}`.
    detect_edges : bool, optional
        Enable or disable edge detection for inversion.
    detect_edges_kwargs : dict, optional
        Dictionary of kwargs used to configure edge detection.
    offset : bool, optional
        Enable or disable offset subtraction. If `True` then the mean value
        of the prior is subtract of the parameter before the rescaling is
        applied.
    """
    def __init__(self, parameters=None, prior_bounds=None, prior=None,
                 rescale_bounds=None, boundary_inversion=None,
                 detect_edges=False, inversion_type='split',
                 detect_edges_kwargs=None, offset=False, update_bounds=True,
                 pre_rescaling=None, post_rescaling=None):

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        self.bounds = None
        self.detect_edge_prime = False

        self.has_pre_rescaling = True
        self.has_post_rescaling = True

        if rescale_bounds is None:
            logger.debug('Using default rescale bounds: [-1, 1]')
            self.rescale_bounds = {p: [-1, 1] for p in self.parameters}
        else:
            if isinstance(rescale_bounds, list):
                self.rescale_bounds = \
                    {p: rescale_bounds for p in self.parameters}
            elif isinstance(rescale_bounds, dict):
                s = set(parameters) - set(rescale_bounds.keys())
                if s:
                    raise RuntimeError(f'Missing bounds for parameters {s}')
                self.rescale_bounds = rescale_bounds
            else:
                raise TypeError(
                    'rescale_bounds must be an instance of list or dict. '
                    f'Got type: {type(rescale_bounds).__name__}')

        if boundary_inversion is not None:
            if isinstance(boundary_inversion, list):
                self.boundary_inversion = \
                    {n: inversion_type for n in boundary_inversion}
            elif isinstance(boundary_inversion, dict):
                self.boundary_inversion = boundary_inversion
            elif isinstance(boundary_inversion, bool):
                self.boundary_inversion = \
                    {p: inversion_type for p in self.parameters}
            else:
                raise TypeError(
                    'boundary_inversion must be an instance of list or dict. '
                    f'Got type: {type(boundary_inversion).__name__}')
        else:
            self.boundary_inversion = []

        for p in self.boundary_inversion:
            self.rescale_bounds[p] = [0, 1]

        self._update_bounds = update_bounds if not detect_edges else True
        self._edges = {n: None for n in self.parameters}
        self.detect_edges = detect_edges
        if self.boundary_inversion:
            self.detect_edges_kwargs = \
                configure_edge_detection(detect_edges_kwargs,
                                         self.detect_edges)

        if self.detect_edges and not self.boundary_inversion:
            raise RuntimeError(
                'Must enable boundary inversion to detect edges')

        if prior == 'uniform':
            self.prior = 'uniform'
            self.has_prime_prior = True
            self._prime_prior = log_uniform_prior
            logger.info(f'Prime prior enabled for {self.name}')
        else:
            self.has_prime_prior = False
            logger.info(f'Prime prior disabled for {self.name}')

        if offset:
            self.offsets = {p: b[0] + np.ptp(b) / 2
                            for p, b in self.prior_bounds.items()}
            logger.debug(f'Offsets: {self.offsets}')
        else:
            self.offsets = {p: 0. for p in self.prior_bounds.keys()}

        self.configure_pre_rescaling(pre_rescaling)
        self.configure_post_rescaling(post_rescaling)

        self.set_bounds(self.prior_bounds)

    def configure_pre_rescaling(self, pre_rescaling):
        if pre_rescaling is not None:
            if isinstance(pre_rescaling, str):
                logger.debug(f'Getting pre-rescaling function {pre_rescaling}')
                self.pre_rescaling, self.pre_rescaling_inv = \
                    rescaling_functions.get(
                        pre_rescaling.lower(), (None, None))
                if self.pre_rescaling is None:
                    raise RuntimeError(
                        f'Unkown rescaling function: {pre_rescaling}')
            elif len(pre_rescaling) == 2:
                self.pre_rescaling = pre_rescaling[0]
                self.pre_rescaling_inv = pre_rescaling[1]
            else:
                raise RuntimeError(
                    'Pre-rescaling must be str or tuple of two functions')
            logger.debug('Disabling prime prior with pre-rescaling')
            self.has_prior_prior = False
            self.has_pre_rescaling = True
        else:
            logger.debug('No pre-rescaling to configure')
            self.has_pre_rescaling = False

    def configure_post_rescaling(self, post_rescaling):

        if post_rescaling is not None:
            if isinstance(post_rescaling, str):
                logger.debug(
                    f'Getting post-rescaling function {post_rescaling}')
                self.post_rescaling, self.post_rescaling_inv = \
                    rescaling_functions.get(
                        post_rescaling.lower(), (None, None))
                if self.post_rescaling is None:
                    raise RuntimeError(
                        f'Unkown rescaling function: {post_rescaling}')
            elif len(post_rescaling) == 2:
                self.post_rescaling = post_rescaling[0]
                self.post_rescaling_inv = post_rescaling[1]
            else:
                raise RuntimeError(
                    'Post-rescaling must be str or tuple of two functions')
            logger.debug('Disabling prime prior with post-rescaling')
            self.has_prior_prior = False

            if post_rescaling == 'logit':
                if self._update_bounds:
                    raise RuntimeError('Cannot use logit with update bounds')
                logger.debug('Setting bounds to [0, 1] for logit')
                self.rescale_bounds = {p: [0, 1] for p in self.parameters}
            self.has_post_rescaling = True
        else:
            logger.debug('No post-rescaling to configure')
            self.has_post_rescaling = False

    def pre_rescaling(self, x):
        """Function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def pre_rescaling_inv(self, x):
        """Inverse of function applied before rescaling to bounds"""
        return x.copy(), np.zeros_like(x)

    def post_rescaling(self, x):
        """Function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def post_rescaling_inv(self, x):
        """Inverse of function applied after rescaling to bounds"""
        return x, np.zeros_like(x)

    def _rescale_to_bounds(self, x, n):
        out = self._rescale_factor[n] * \
                ((x - self.bounds[n][0]) /
                 (self.bounds[n][1] - self.bounds[n][0])) \
                + self._rescale_shift[n]

        log_j = (-np.log(self.bounds[n][1] - self.bounds[n][0])
                 + np.log(self._rescale_factor[n]))
        return out, log_j

    def _inverse_rescale_to_bounds(self, x, n):
        out = (self.bounds[n][1] - self.bounds[n][0]) \
               * (x - self._rescale_shift[n]) \
               / self._rescale_factor[n] + self.bounds[n][0]

        log_j = (np.log(self.bounds[n][1] - self.bounds[n][0])
                 - np.log(self._rescale_factor[n]))

        return out, log_j

    def _apply_inversion(self, x, x_prime, log_j, p, pp, compute_radius,
                         test=None):
        if self._edges[p] is None:
            self._edges[p] = detect_edge(x_prime[pp], test=test,
                                         **self.detect_edges_kwargs)
            self.update_prime_prior_bounds()

        if self._edges[p]:
            logger.debug(f'Apply inversion for {p} to {self._edges[p]} bound')
            logger.debug('Fixing bounds to [0, 1]')
            logger.debug('Rescaling')
            x_prime[pp], lj = \
                rescale_zero_to_one(x_prime[pp] - self.offsets[p],
                                    *self.bounds[p])
            log_j += lj
            if self._edges[p] == 'upper':
                x_prime[pp] = 1 - x_prime[pp]
            if (self.boundary_inversion[p] == 'duplicate' or compute_radius):
                logger.debug('Inverting with duplication')
                x_inv = x_prime.copy()
                x_inv[pp] *= -1
                x_prime = np.concatenate([x_prime, x_inv])
                x = np.concatenate([x,  x])
                log_j = np.concatenate([log_j, log_j])
            else:
                logger.debug('Inverting with splitting')
                inv = np.random.choice(x_prime.size,
                                       x_prime.size // 2,
                                       replace=False)
                x_prime[pp][inv] *= -1
        else:
            logger.debug(f'Not using inversion for {p}')
            logger.debug(f'Rescaling to {self.rescale_bounds[p]}')
            x_prime[pp], lj = \
                rescale_minus_one_to_one(x_prime[pp] - self.offsets[p],
                                         xmin=self.bounds[p][0],
                                         xmax=self.bounds[p][1])

            log_j += lj

        return x, x_prime, log_j

    def _reverse_inversion(self, x, x_prime, log_j, p, pp):
        if self._edges[p]:
            inv = x[p] < 0.
            x[p][~inv] = x[p][~inv]
            x[p][inv] = -x[p][inv]

            if self._edges[p] == 'upper':
                x[p] = 1 - x[p]
            x[p], lj = inverse_rescale_zero_to_one(
                x[p], *self.bounds[p])
            x[p] += self.offsets[p]
            log_j += lj
        else:
            x[p], lj = inverse_rescale_minus_one_to_one(
                x[p], xmin=self.bounds[p][0], xmax=self.bounds[p][1])
            x[p] += self.offsets[p]
            log_j += lj
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, compute_radius=False,
                       **kwargs):
        """
        Rescale inputs to the prime space

        Parameters
        ----------
        x, x_prime :  array_like
            Arrays of samples in the physical and prime space
        log_j : array_like
            Array of values of log-Jacboian
        compute_radius : bool, optional
            If true force duplicate for inversion
        kwargs :
            Parsed to inversion function
        """
        for p, pp in zip(self.parameters, self.prime_parameters):
            if self.has_pre_rescaling:
                x_prime[pp], lj = self.pre_rescaling(x[p])
                log_j += lj
            else:
                x_prime[pp] = x[p].copy()

            if p in self.boundary_inversion:
                x, x_prime, log_j = self._apply_inversion(
                        x, x_prime, log_j, p, pp, compute_radius, **kwargs)
            else:
                x_prime[pp], lj = \
                    self._rescale_to_bounds(x[p] - self.offsets[p], p)
                log_j += lj
            if self.has_post_rescaling:
                x_prime[pp], lj = self.post_rescaling(x_prime[pp])
                log_j += lj
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Map inputs to the physical space from the prime space"""
        for p, pp in zip(reversed(self.parameters),
                         reversed(self.prime_parameters)):
            if self.has_post_rescaling:
                x[p], lj = self.post_rescaling_inv(x_prime[pp])
                log_j += lj
            else:
                x[p] = x_prime[pp].copy()
            if p in self.boundary_inversion:
                x, x_prime, log_j = self._reverse_inversion(
                    x, x_prime, log_j, p, pp, **kwargs)
            else:
                x[p], lj = self._inverse_rescale_to_bounds(x[p], p)
                x[p] += self.offsets[p]
                log_j += lj
            if self.has_pre_rescaling:
                x[p], lj = self.pre_rescaling_inv(x[p])
                log_j += lj
        return x, x_prime, log_j

    def reset_inversion(self):
        """Reset the edges for inversion"""
        self._edges = {n: None for n in self.parameters}

    def set_bounds(self, prior_bounds):
        """Set the initial bounds for rescaling"""
        self._rescale_factor = \
            {p: np.ptp(self.rescale_bounds[p]) for p in self.parameters}
        self._rescale_shift = \
            {p: self.rescale_bounds[p][0] for p in self.parameters}

        self.prior_bounds = \
            {p: self.pre_rescaling(prior_bounds[p])[0]
             for p in self.parameters}
        self.bounds = {p: self.pre_rescaling(b - self.offsets[p])[0]
                       for p, b in prior_bounds.items()}
        logger.debug(f'Initial bounds: {self.bounds}')
        self.update_prime_prior_bounds()

    def update_bounds(self, x):
        """Update the bounds used for the reparameterisation"""
        if self._update_bounds:
            self.bounds = \
                {p: [self.pre_rescaling(np.min(x[p]))[0] - self.offsets[p],
                     self.pre_rescaling(np.max(x[p]))[0] - self.offsets[p]]
                 for p in self.parameters}
            logger.debug(f'New bounds: {self.bounds}')
            self.update_prime_prior_bounds()

        else:
            logger.debug('Update bounds not enabled')

    def update_prime_prior_bounds(self):
        """Update the prior bounds used for the prime prior"""
        if self.has_prime_prior:
            self.prime_prior_bounds = \
                {pp: self.post_rescaling(np.asarray(determine_rescaled_bounds(
                    self.prior_bounds[p][0], self.prior_bounds[p][1],
                    self.bounds[p][0], self.bounds[p][1], self._edges[p],
                    self.offsets[p], rescale_bounds=self.rescale_bounds[p],
                    inversion=p in self.boundary_inversion)))[0]
                 for p, pp in zip(self.parameters, self.prime_parameters)}
            logger.debug(f'New prime bounds: {self.prime_prior_bounds}')

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            log_p = 0
            for pp in self.prime_parameters:
                log_p += self._prime_prior(x_prime[pp],
                                           *self.prime_prior_bounds[pp])
            return log_p
        else:
            raise RuntimeError(
                f'Prime prior is not configured for {self.name}')


class Angle(Reparameterisation):
    """Reparameterisation for a single angle.

    This reparameterisations converts an angle to Cartesian coordinates
    using either a corresponding radial parameter or an auxiliary radial
    parameter. When using the auxiliary parameter, samples are drawn from
    a chi-distribution with two degrees of freedom.

    Parameters
    ----------
    parameters : str or list
        Parameter(s) to use for the reparameterisation. Either just an angle
        or an angle and corresponding radial parameter
    prior_bounds : dict
        Dictionary of prior bounds. Does not need to be specified when defining
        `reparameterisations`.
    scale : float, optional
        Value used to rescale the angle before converting to Cartesian
        coordiantes.
    prior : {'uniform', 'sine', None}
        Type of prior being used for sampling this angle. If specified, the
        prime prior is enabled. If None then it is disabled.
    """
    def __init__(self, parameters=None, prior_bounds=None, scale=1.0,
                 prior=None):
        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        if len(self.parameters) == 1:
            self.parameters.append(self.parameters[0] + '_radial')
            self.chi = stats.chi(2)
            self.has_prior = True
        elif len(self.parameters) == 2:
            self.chi = False
        else:
            raise RuntimeError('Too many parameters for Angle')

        self.scale = scale

        if prior_bounds[self.angle][0] == 0:
            self._zero_bound = True
        else:
            self._zero_bound = False

        self.prime_parameters = [self.angle + '_x', self.angle + '_y']
        self.requires = []

        if prior in ['uniform', 'sine']:
            self.prior = prior
            self.has_prime_prior = True
            if self.prior == 'uniform':
                self._prime_prior = log_2d_cartesian_prior
                self._k = self.scale * np.pi
            else:
                self._prime_prior = log_2d_cartesian_prior_sine
                self._k = np.pi
            logger.info(f'Prime prior enabled for {self.name}')
        else:
            self.has_prime_prior = False
            logger.info(f'Prime prior disabled for {self.name}')

    @property
    def angle(self):
        """The name of the angular parameter"""
        return self.parameters[0]

    @property
    def radial(self):
        """The name of the radial parameter"""
        return self.parameters[1]

    @property
    def radius(self):
        """The name of the radial parameter (equivalent to radial)"""
        return self.parameters[1]

    @property
    def x(self):
        """The name of x coordinate"""
        return self.prime_parameters[0]

    @property
    def y(self):
        """The name of y coordinate"""
        return self.prime_parameters[1]

    def _rescale_radial(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[1]], x, x_prime, log_j

    def _rescale_angle(self, x, x_prime, log_j, **kwargs):
        return x[self.parameters[0]] * self.scale, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert the angle to Cartesian coordinates"""
        angle, x, x_prime, log_j = self._rescale_angle(
            x, x_prime, log_j, **kwargs)

        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r, x, x_prime, log_j = self._rescale_radial(
                x, x_prime, log_j, **kwargs)
        if any(r < 0):
            raise RuntimeError('Radius cannot be negative.')

        x_prime[self.prime_parameters[0]] = r * np.cos(angle)
        x_prime[self.prime_parameters[1]] = r * np.sin(angle)
        log_j += np.log(r)
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to an angle and radial component"""
        x[self.parameters[1]] = \
            np.sqrt(x_prime[self.prime_parameters[0]] ** 2
                    + x_prime[self.prime_parameters[1]] ** 2)
        if self._zero_bound:
            x[self.parameters[0]] = \
                np.arctan2(x_prime[self.prime_parameters[1]],
                           x_prime[self.prime_parameters[0]]) % \
                (2. * np.pi) / self.scale
        else:
            x[self.parameters[0]] = \
                np.arctan2(x_prime[self.prime_parameters[1]],
                           x_prime[self.prime_parameters[0]]) / self.scale

        log_j -= np.log(x[self.parameters[1]])
        x, x_prime, log_j = self._inverse_rescale_angle(x, x_prime, log_j)

        return x, x_prime, log_j

    def log_prior(self, x):
        """Prior for radial parameter"""
        return self.chi.logpdf(x[self.parameters[1]])

    def x_prime_log_prior(self, x_prime):
        """Compute the prior in the prime space assuming a uniform prior"""
        if self.has_prime_prior:
            return self._prime_prior(
                x_prime[self.prime_parameters[0]],
                x_prime[self.prime_parameters[1]],
                k=self._k)
        else:
            raise RuntimeError('Prime prior')


class ToCartesian(Angle):
    """Convert a paraemter to Cartesian coordinates"""
    def __init__(self, mode='split', scale=np.pi, **kwargs):
        super().__init__(scale=np.pi, **kwargs)

        self.mode = mode
        if self.mode not in ['duplicate', 'split', 'half']:
            raise RuntimeError(f'Unknown mode: {self.mode}')
        logger.debug(f'Using mode: {self.mode}')

        self._zero_bound = False
        self._k = self.prior_bounds[self.parameters[0]][1]

    def _rescale_angle(self, x, x_prime, log_j, compute_radius=False,
                       **kwargs):
        angle, lj = rescale_zero_to_one(
            x[self.parameters[0]], *self.prior_bounds[self.parameters[0]])
        log_j += lj
        if self.mode == 'duplicate' or compute_radius:
            angle = np.concatenate([angle, -angle])
            x = np.concatenate([x, x])
            x_prime = np.concatenate([x_prime, x_prime])
            log_j = np.concatenate([log_j, log_j])
        elif self.mode == 'split':
            neg = np.random.choice(angle.size, angle.size // 2, replace=False)
            angle[neg] *= -1

        angle *= self.scale
        return angle, x, x_prime, log_j

    def _inverse_rescale_angle(self, x, x_prime, log_j):
        x[self.angle], lj = inverse_rescale_zero_to_one(
            np.abs(x[self.parameters[0]]),
            *self.prior_bounds[self.parameters[0]])
        log_j += lj
        return x, x_prime, log_j


class AnglePair(Reparameterisation):
    """Reparameterisation for a pair of angles and a radial component"""
    def __init__(self, parameters=None, prior_bounds=None,
                 prior=None, convention=None):

        self._conventions = ['az-zen', 'ra-dec']

        if len(parameters) not in [2, 3]:
            raise RuntimeError(
                'Must use a pair of angles or a pair plus a radius')

        super().__init__(parameters=parameters, prior_bounds=prior_bounds)

        # Determine which parameter is for the horizontal plane
        # and which is for the vertical plane
        logger.debug('Checking order of parameters')
        b = np.ptp([prior_bounds[p] for p in parameters], axis=1)
        hz = np.where(b == (2 * np.pi))[0][0]
        vt = np.where(b == np.pi)[0][0]

        if len(parameters) == 3:
            r = list({0, 1, 2} - {hz, vt})[0]
            parameters[0], parameters[1], parameters[2] = \
                parameters[hz], parameters[vt], parameters[r]
        else:
            parameters[0], parameters[1] = parameters[hz], parameters[vt]

        if len(parameters) == 2:
            m = '_'.join(parameters)
            parameters.append(f'{m}_radial')
            self.chi = stats.chi(3)
            self.has_prior = True
        else:
            m = '_'.join(parameters)
            self.chi = False

        logger.debug(f'Parameters are: {parameters}')

        self.parameters = parameters
        self.prime_parameters = [f'{m}_{x}' for x in ['x', 'y', 'z']]

        print(prior)
        if prior == 'isotropic' and self.chi:
            self.has_prime_prior = True
            logger.info(f'Prime prior enabled for {self.name}')
        else:
            self.has_prime_prior = False
            logger.info(f'Prime prior disabled for {self.name}')

        if convention is None:
            if (self.prior_bounds[self.parameters[1]][0] == 0 and
                    self.prior_bounds[self.parameters[1]][1] == np.pi):
                self.convention = 'az-zen'
            elif (self.prior_bounds[self.parameters[1]][0] == -np.pi / 2 and
                    self.prior_bounds[self.parameters[1]][1] == np.pi / 2):
                self.convention = 'ra-dec'
            else:
                raise RuntimeError
        elif convention in self._conventions:
            self.convention = convention
        else:
            raise RuntimeError(f'Unknown convention: {convention}')

        logger.debug(f'Using convention: {self.convention}')

        self.requires = []

    @property
    def angles(self):
        return self.parameters[:2]

    @property
    def radial(self):
        return self.parameters[-1]

    @property
    def x(self):
        return self.prime_parameters[0]

    @property
    def y(self):
        return self.prime_parameters[1]

    @property
    def z(self):
        return self.prime_parameters[2]

    def _az_zen(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = \
            r * np.sin(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        x_prime[self.prime_parameters[1]] = \
            r * np.sin(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        x_prime[self.prime_parameters[2]] = \
            r * np.cos(x[self.parameters[1]])
        log_j += (2 * np.log(r) + np.log(np.sin(x[self.parameters[1]])))
        return x, x_prime, log_j

    def _ra_dec(self, x, x_prime, log_j, r):
        x_prime[self.prime_parameters[0]] = \
            r * np.cos(x[self.parameters[1]]) * np.cos(x[self.parameters[0]])
        x_prime[self.prime_parameters[1]] = \
            r * np.cos(x[self.parameters[1]]) * np.sin(x[self.parameters[0]])
        x_prime[self.prime_parameters[2]] = \
            r * np.sin(x[self.parameters[1]])
        log_j += (2 * np.log(r) + np.log(np.cos(x[self.parameters[1]])))
        return x, x_prime, log_j

    def _inv_az_zen(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2. +
            x_prime[self.prime_parameters[1]] ** 2. +
            x_prime[self.prime_parameters[2]] ** 2.)
        x[self.parameters[0]] = \
            np.arctan2(x_prime[self.prime_parameters[1]],
                       x_prime[self.prime_parameters[0]]) % (2. * np.pi)
        x[self.parameters[1]] = \
            np.arctan2(np.sqrt(x_prime[self.prime_parameters[0]] ** 2.
                               + x_prime[self.prime_parameters[1]] ** 2.),
                       x_prime[self.prime_parameters[2]])
        log_j += (-2 * np.log(x[self.parameters[2]])
                  - np.log(np.sin(x[self.parameters[1]])))

        return x, x_prime, log_j

    def _inv_ra_dec(self, x, x_prime, log_j):
        x[self.parameters[2]] = np.sqrt(
            x_prime[self.prime_parameters[0]] ** 2. +
            x_prime[self.prime_parameters[1]] ** 2. +
            x_prime[self.prime_parameters[2]] ** 2.)
        x[self.parameters[0]] = \
            np.arctan2(x_prime[self.prime_parameters[1]],
                       x_prime[self.prime_parameters[0]]) % (2. * np.pi)
        x[self.parameters[1]] = \
            np.arctan2(x_prime[self.prime_parameters[2]],
                       np.sqrt(x_prime[self.prime_parameters[0]] ** 2.
                               + x_prime[self.prime_parameters[1]] ** 2.))

        log_j += (-2 * np.log(x[self.parameters[2]])
                  - np.log(np.cos(x[self.parameters[1]])))
        return x, x_prime, log_j

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Convert the spherical polar angles to Cartesian coordinates
        """
        if self.chi:
            r = self.chi.rvs(size=x.size)
        else:
            r = x[self.radial]
        if any(r < 0):
            raise RuntimeError('Radius cannot be negative.')

        if self.convention == 'az-zen':
            return self._az_zen(x, x_prime, log_j, r)
        else:
            return self._ra_dec(x, x_prime, log_j, r)

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """Convert from Cartesian to spherical polar angles"""
        if self.convention == 'az-zen':
            return self._inv_az_zen(x, x_prime, log_j)
        else:
            return self._inv_ra_dec(x, x_prime, log_j)
        return x, x_prime, log_j

    def log_prior(self, x):
        """Prior for radial parameter"""
        return self.chi.logpdf(x[self.parameters[2]])

    def x_prime_log_prior(self, x_prime):
        """
        Log probability of 3d Cartesian coordinates for an isotropic
        distribution of angles and a radial component drawn from a chi
        distribution with three degrees of freedom.
        """
        if self.has_prime_prior:
            return - 1.5 * np.log(2 * np.pi) \
                   - np.sum([x_prime[pp] ** 2 for pp in self.prime_parameters],
                            axis=0) / 2
        else:
            raise RuntimeError('x prime prior')


default_reparameterisations = {
    'default': (RescaleToBounds, None),
    'rescaletobounds': (RescaleToBounds, None),
    'rescale-to-bounds': (RescaleToBounds, None),
    'offset': (RescaleToBounds, {'offset': True}),
    'inversion': (RescaleToBounds, {'detect_edges': True,
                                    'boundary_inversion': True,
                                    'inversion_type': 'split'}),
    'inversion-duplicate': (RescaleToBounds, {'detect_edges': True,
                                              'boundary_inversion': True,
                                              'inversion_type': 'duplicate'}),
    'angle': (Angle, {}),
    'angle-pi': (Angle, {'scale': 2.0, 'prior': 'uniform'}),
    'angle-2pi': (Angle, {'scale': 1.0, 'prior': 'uniform'}),
    'angle-sine': (Angle, {'scale': 1.0, 'prior': 'sine'}),
    'angle-pair': (AnglePair, None),
    'to-cartesian': (ToCartesian, None),
    'none': (NullReparameterisation, None),
    None: (NullReparameterisation, None),
}



#!/usr/bin/env python

from scipy.stats import norm, halfnorm
from scipy.special import xlogy

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = './outdir/half_gaussian/'
logger = setup_logger(output=output, log_level='DEBUG')


class HalfGaussian(Model):
    """
    Two-dimensional Guassian with a bound at y=0
    """
    def __init__(self):

        self.names = ['x', 'y']
        self.bounds = {'x': [-10, 10], 'y': [0, 10]}

    def log_prior(self, x):
        """
        Log prior that accepts structured arrays with fields:
        x, y, logP, logL
        """
        log_p = 0.
        for i, n in enumerate(self.names):
            log_p += xlogy(1, (x[n] >= self.bounds[n][0])
                           & (x[n] <= self.bounds[n][1]))
            - xlogy(1, self.bounds[n][1] - self.bounds[n][0])
        return log_p

    def log_likelihood(self, x):
        """
        Log likelihood that accepts structured arrays of length
        one with fields:
        x, y, logP, logL
        """
        log_l = 0
        log_l += norm.logpdf(x['x'])
        log_l += halfnorm.logpdf(x['y'])
        return log_l


# Configure the flow
flow_config = dict(
        max_epochs=50,
        patience=10,
        model_config=dict(n_blocks=2, n_neurons=2, n_layers=2)
        )

# Enable boundary inversion for 'y' since we expect the posterior to rail
# against the bounds
fp = FlowSampler(HalfGaussian(), output=output, resume=False, nlive=1000,
                 plot=True, flow_config=flow_config, training_frequency=None,
                 maximum_uninformed=1000, rescale_parameters=True, seed=1234,
                 boundary_inversion=['y'], detect_edges=True)

# And go!
fp.run()



#!/usr/bin/env python

"""
Example of running nessai with bilby on a gravitational wave likelihood. This
example should take ~20 minutes to run.

Based on the Bilby example: https://git.ligo.org/lscsoft/bilby
"""
import bilby
import numpy as np

outdir = './outdir/'
label = 'gw_example'

bilby.core.utils.setup_logger(outdir=outdir, label=label, log_level='INFO')

duration = 4.
sampling_frequency = 2048.

np.random.seed(151226)

# Use an injection that is imilar to GW150914
injection_parameters = dict(
    total_mass=66., mass_ratio=0.9, a_1=0.4, a_2=0.3, tilt_1=0.5, tilt_2=1.0,
    phi_12=1.7, phi_jl=0.3, luminosity_distance=2000, theta_jn=0.4, psi=2.659,
    phase=1.3, geocent_time=1126259642.413, ra=1.375, dec=-1.2108)

waveform_arguments = dict(waveform_approximant='IMRPhenomPv2',
                          reference_frequency=50.)

# Create the waveform_generator
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    sampling_frequency=sampling_frequency,
    duration=duration,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=(bilby.gw.conversion
                          .convert_to_lal_binary_black_hole_parameters),
    waveform_arguments=waveform_arguments)

# Set up interferometers
ifos = bilby.gw.detector.InterferometerList(['H1', 'L1', 'V1'])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency, duration=duration,
    start_time=injection_parameters['geocent_time'] - 3)
ifos.inject_signal(waveform_generator=waveform_generator,
                   parameters=injection_parameters)

# Set up prior
# Nessai is designed to sample mass ratio and chirp mass
priors = bilby.gw.prior.BBHPriorDict()
priors['chirp_mass'] = bilby.prior.Uniform(
    name='chirp_mass', latex_label='$m_c$', minimum=13, maximum=45,
    unit='$M_{\\odot}$')
priors['mass_ratio'] = bilby.prior.Uniform(
    name='mass_ratio', latex_label='q', minimum=0.125, maximum=1.0)

priors['geocent_time'] = bilby.core.prior.Uniform(
    minimum=injection_parameters['geocent_time'] - 0.1,
    maximum=injection_parameters['geocent_time'] + 0.1,
    name='geocent_time', latex_label='$t_c$', unit='$s$')

# Only sample a subset of the parameters
for key in ['a_1', 'a_2', 'tilt_1', 'tilt_2', 'phi_12', 'phi_jl',
            'luminosity_distance', 'psi', 'geocent_time', 'ra', 'dec']:
    priors[key] = injection_parameters[key]

# Initialise the likelihood
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator,
    phase_marginalization=True, priors=priors)

# Configuration for the normalising flow
flow_config = {
    "lr": 0.001,
    "batch_size": 1000,
    "val_size": 0.1,
    "max_epochs": 200,
    "patience": 50,
    "model_config": {
        "n_blocks": 2,
        "n_neurons": 4,
        "n_layers": 2,
        "ftype": "realnvp",
        "kwargs": {
            "batch_norm_between_layers": True,
            "linear_transform": "lu"
        }
    }
}

# Run sampler

# The `flow_class` should be set to `GWFlowProposal` for GW PE. This includes
# specific default reparameterisations for certain parameters. For example,
# it knows that theta_jn is angle with a sine prior.

# The dictionary 'reparameterisations' tells nessai what to do with each
# parameter. It has default behaviour for most of standard parameters
# for BBHs in Bilby. You can override any of these.
# In this example we've specified the prior used for each parameter, this
# is not necessary but allows the sampler to use priors defined in the
# prime space instead of the non-prime space. This is much faster since
# these priors a simpler and do not have the overhead the Bilby priors have.

# See the documentation for list of available reparameterisations
result = bilby.core.sampler.run_sampler(
    likelihood=likelihood,
    priors=priors,
    outdir=outdir,
    injection_parameters=injection_parameters,
    label=label,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
    flow_class='GWFlowProposal',
    sampler='nessai',
    resume=False,
    plot=True,
    nlive=2000,
    maximum_uninformed=4000,
    flow_config=flow_config,
    seed=150914,
    analytic_priors=True,      # Bilby priors can be sampled from directly
    reparameterisations={
        'chirp_mass': {'reparameterisation': 'mass', 'prior': 'uniform'},
        'mass_ratio': {'reparameterisation': 'mass_ratio', 'prior': 'uniform'},
        'theta_jn': {'reparameterisation': 'angle-sine', 'prior': 'sine'},
        }
    )

# Produce corner plots
result.plot_corner()



"""
Test the prior functions.
"""
import numpy as np
import pytest

from nessai import priors


def test_uniform_in_bounds():
    """Test behaviour if x is out of bounds"""
    assert priors.log_uniform_prior(0) == 0.


def test_uniform_out_of_bounds():
    """Test behaviour if x is out of bounds"""
    assert priors.log_uniform_prior(-2) == -np.inf


@pytest.mark.parametrize('k', [np.pi, 2 * np.pi])
def test_2d_cartesian_origin(k):
    """Test the value of the 2d Cartesian prior at the origin"""
    assert priors.log_2d_cartesian_prior(0, 0, k=k) == -np.log(k)


def test_2d_cartesian_sine_negative():
    """Test to ensure prior is 0 if y < 0"""
    x = np.array([[0.5], [-1.0]])
    assert priors.log_2d_cartesian_prior_sine(x[0], x[1]) == -np.inf


def test_2d_cartesian_sine_in_valid_range():
    """Test to ensure prior raises a runtime error if k != pi"""
    x = np.array([[0.5], [1.0]])
    with pytest.raises(RuntimeError) as excinfo:
        priors.log_2d_cartesian_prior_sine(x[0], x[1], k=2 * np.pi)
    assert 'incompatible' in str(excinfo.value)


def test_3d_cartesian_at_origin():
    """Test the value of 3d Cartesian prior at the origin"""
    assert priors.log_3d_cartesian_prior(0, 0, 0) == (-1.5 * np.log(2 * np.pi))



# -*- coding: utf-8 -*-
"""
Test the base flow class
"""
import pytest

from nessai.flows import BaseFlow


@pytest.mark.parametrize('method', ['forward', 'inverse', 'sample', 'log_prob',
                                    'base_distribution_log_prob',
                                    'forward_and_log_prob',
                                    'sample_and_log_prob'])
def test_base_flow_methods(method):
    """Test to make sure all of the method in the base class raise errors"""
    m = getattr(BaseFlow(), method)
    assert m
    with pytest.raises(NotImplementedError):
        m(None)




import numpy as np
import pytest

from nessai.flowmodel import update_config, FlowModel


@pytest.fixture()
def data_dim():
    return 2


@pytest.fixture(scope='function')
def flow_model(flow_config, data_dim, tmpdir):
    flow_config['model_config']['n_inputs'] = data_dim
    output = str(tmpdir.mkdir('flowmodel'))
    return FlowModel(flow_config, output=output)


def test_update_config_none():
    """Test update config for no input"""
    config = update_config(None)
    assert 'model_config' in config


def test_update_config_invalid_type():
    """Test update config when an invald argument is specified"""
    with pytest.raises(TypeError) as excinfo:
        update_config(False)
    assert 'Must pass a dictionary' in str(excinfo.value)


def test_init_no_config(tmpdir):
    """Test the init method with no config specified"""
    output = str(tmpdir.mkdir('no_config'))
    default_config = update_config(None)
    fm = FlowModel(output=output)

    assert fm.model_config == default_config['model_config']


def test_init_config_class(tmpdir):
    """Test the init and save methods when specifying `flow` as a class"""
    from nessai.flows import FlexibleRealNVP
    output = str(tmpdir.mkdir('no_config'))
    config = dict(model_config=dict(flow=FlexibleRealNVP))
    fm = FlowModel(config=config, output=output)

    assert fm.model_config['flow'].__name__ == 'FlexibleRealNVP'


@pytest.mark.parametrize('val_size, batch_size', [(0.1, 1),
                                                  (0.1, 100),
                                                  (0.1, 'all'),
                                                  (0.5, 'all')])
def test_prep_data(flow_model, data_dim, val_size, batch_size):
    """
    Test the data prep, make sure batch sizes and validation size
    produce the correct result.
    """
    n = 1000
    x = np.random.randn(n, data_dim)

    train, val = flow_model.prep_data(x, val_size, batch_size)
    train_batch = next(iter(train))[0]
    val_batch = next(iter(val))[0]

    if batch_size == 'all':
        batch_size = int(n * (1 - val_size))

    assert list(train_batch.shape) == [batch_size, data_dim]
    assert list(val_batch.shape) == [int(val_size * n), data_dim]
    assert len(train) * train_batch.shape[0] + val_batch.shape[0] == n


@pytest.mark.parametrize('batch_size', [None, '10', True, False])
def test_incorrect_batch_size_type(flow_model, data_dim, batch_size):
    """Ensure the non-interger batch sizes do not work"""
    n = 1000
    x = np.random.randn(n, data_dim)
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.prep_data(x, 0.1, batch_size)
    assert 'Unknown batch size' in str(excinfo.value)


def test_training(flow_model, data_dim):
    """Test class until training"""
    x = np.random.randn(1000, data_dim)
    flow_model.train(x)
    assert flow_model.weights_file is not None


@pytest.mark.parametrize('key, value', [('anneling', True),
                                        ('noise_scale', 0.1),
                                        ('noise_scale', 'adaptive'),
                                        ('max_epochs', 51)])
def test_training_additional_config_args(flow_config, data_dim, tmpdir,
                                         key, value):
    """
    Test training with different config args
    """
    flow_config['model_config']['n_inputs'] = data_dim
    flow_config[key] = value

    output = str(tmpdir.mkdir('flowmodel'))
    flow_model = FlowModel(flow_config, output=output)

    assert getattr(flow_model, key) == value

    x = np.random.randn(1000, data_dim)
    flow_model.train(x)


def test_early_optimiser_init(flow_model):
    """Ensure calling the opitmiser before the model raises an error"""
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.get_optimiser()
    assert 'Cannot initialise optimiser' in str(excinfo.value)


def test_sample_and_log_prob_not_initialised(flow_model, data_dim):
    """
    Ensure user cannot call the method before the model initialise.
    """
    with pytest.raises(RuntimeError) as excinfo:
        flow_model.sample_and_log_prob()
    assert 'Model is not initialised' in str(excinfo.value)


@pytest.mark.parametrize('N', [1, 100, 1000])
def test_sample_and_log_prob_no_latent(flow_model, data_dim, N):
    """
    Test the basic use of sample and log prob and ensure correct output
    shapes.
    """
    flow_model.initialise()
    x, log_prob = flow_model.sample_and_log_prob(N=N)
    assert x.shape == (N, data_dim)
    assert log_prob.size == N



import logging

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:   # for Python < 3.8
    from importlib_metadata import version, PackageNotFoundError


try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    pass

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = ['flowsampler'
           'nestedsampler'
           'proposal',
           'model'
           'flowmodel',
           'flows',
           'utils']



#!/usr/bin/env python

# Example of using Nessai with Bilby (Requires seperate installation)
# See 2d_gaussian.py for a more detailed explanation of using nessai

import bilby
import numpy as np

# The output from the sampler will be saved to:
# '$outdir/$label_flowproposal/'
# alongside the usual bilby outputs
outdir = './outdir/'
label = 'bilby_example'

# Setup the bilby logger
bilby.core.utils.setup_logger(outdir=outdir, label=label, log_level='INFO')

# Define a likelihood using Bilby


class SimpleGaussianLikelihood(bilby.Likelihood):

    def __init__(self):
        """
        A very simple Gaussian likelihood
        """
        super().__init__(parameters={'x': None, 'y': None})

    def log_likelihood(self):
        x = self.parameters['x']
        y = self.parameters['y']
        return -0.5*(x ** 2. + y ** 2.) - np.log(2.0 * np.pi)


# Define priors (this provides the bounds that are then used in Nessai)
priors = dict(x=bilby.core.prior.Uniform(-10, 10, 'x'),
              y=bilby.core.prior.Uniform(-10, 10, 'y'))

# Instantiate the likleihood
likelihood = SimpleGaussianLikelihood()

# Configure the normalisng flow
flow_config = dict(
        max_epochs=50,
        patience=10,
        model_config=dict(n_blocks=2, n_neurons=4, n_layers=2,
                          device_tag='cpu',
                          kwargs=dict(batch_norm_between_layers=True))
        )

# Run using bilby.run_sampler, any kwargs are parsed to the sampler
# NOTE: when using Bilby if the priors can be sampled analytically  the flag
# `analytic_priors` enables faster initial sampling. See 'further details' in
# the documentation for more details
result = bilby.run_sampler(outdir=outdir, label=label, resume=False, plot=True,
                           likelihood=likelihood, priors=priors,
                           sampler='nessai', nlive=1000,
                           maximum_uninformed=2000, flow_config=flow_config,
                           rescale_parameters=True,
                           injection_parameters={'x': 0.0, 'y': 0.0},
                           analytic_priors=True, seed=1234)



"""
Test the distance converter classes.
"""
import numpy as np
import pytest
from scipy import stats

from nessai.gw.utils import (
    PowerLawConverter,
    )


@pytest.mark.parametrize('power', [1, 2, 3, 4])
@pytest.mark.flaky(run=5)
def test_power_law_converter_distribution(power):
    """
    Check that the distribution of resulting samples is uniform when
    converting from a power law.
    """
    c = PowerLawConverter(power, scale=1)
    x = stats.powerlaw(power + 1).rvs(size=10000)
    y, _ = c.to_uniform_parameter(x)
    d, p = stats.kstest(y, 'uniform')
    assert p >= 0.05


@pytest.mark.parametrize('power', [1, 2, 3, 4])
def test_power_law_converter_inversion(power):
    """
    Check that the power law inversion is invertible
    """
    c = PowerLawConverter(power, scale=1)
    x = stats.powerlaw(power + 1).rvs(size=1000)
    y, _ = c.to_uniform_parameter(x)
    x_out, _ = c.from_uniform_parameter(y)
    np.testing.assert_array_almost_equal(x, x_out)



import logging
import os

import numpy as np
from scipy import stats
from scipy.special import logsumexp

from ..flowmodel import FlowModel
from ..livepoint import (
    live_points_to_array,
    )

from ..plot import plot_live_points
from .flowproposal import FlowProposal

logger = logging.getLogger(__name__)


class AugmentedFlowProposal(FlowProposal):

    def __init__(self, model, augment_features=1, **kwargs):
        super().__init__(model, **kwargs)
        self.augment_features = augment_features

    def set_rescaling(self):
        super().set_rescaling()
        self.augment_names = [f'e_{i}' for i in range(self.augment_features)]
        self.names += self.augment_names
        self.rescaled_names += self.augment_names
        logger.info(f'augmented x space parameters: {self.names}')
        logger.info(f'parameters to rescale {self.rescale_parameters}')
        logger.info(
                f'augmented x prime space parameters: {self.rescaled_names}'
                )

    def initialise(self):
        """
        Initialise the proposal class
        """
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        self.set_rescaling()

        m = np.ones(self.rescaled_dims)
        m[-self.augment_features:] = -1
        self.flow_config['model_config']['kwargs']['mask'] = m

        self.flow_config['model_config']['n_inputs'] = self.rescaled_dims

        self.flow = FlowModel(config=self.flow_config, output=self.output)
        self.flow.initialise()
        self.populated = False
        self.initialised = True

    def _rescale_with_bounds(self, x, generate_augment=True):
        """
        Rescale the inputs to [-1, 1] using the bounds as the min and max
        """
        x_prime = np.zeros([x.size], dtype=self.x_prime_dtype)
        log_J = 0.
        for n, rn in zip(self.names, self.rescaled_names):
            if n in self.rescale_parameters:
                x_prime[rn] = 2 * ((x[n] - self.model.bounds[n][0])
                                   / (self.model.bounds[n][1]
                                      - self.model.bounds[n][0])) - 1

                log_J += np.log(2) - np.log(self.model.bounds[n][1]
                                            - self.model.bounds[n][0])
            elif n not in self.augment_names:
                x_prime[rn] = x[n]
        x_prime['logP'] = x['logP']
        x_prime['logL'] = x['logL']
        if generate_augment:
            if x_prime.size == 1:
                x_prime[self.augment_names] = self.augment_features * (0.,)
            else:
                for an in self.augment_names:
                    x_prime[an] = np.random.randn(x_prime.size)

        return x_prime, log_J

    def _inverse_rescale_with_bounds(self, x_prime):
        """
        Rescale the inputs from the prime space to the phyiscal space
        using the model bounds
        """
        x = np.zeros([x_prime.size], dtype=self.x_dtype)
        log_J = 0.
        for n, rn in zip(self.names, self.rescaled_names):
            if n in self.rescale_parameters:
                x[n] = (self.model.bounds[n][1] - self.model.bounds[n][0]) \
                        * ((x_prime[rn] + 1) / 2) + self.model.bounds[n][0]
                log_J += np.log(self.model.bounds[n][1]
                                - self.model.bounds[n][0]) - np.log(2)
            else:
                x[n] = x_prime[rn]
        x['logP'] = x_prime['logP']
        x['logL'] = x_prime['logL']
        return x, log_J

    def augmented_prior(self, x):
        """
        Log guassian for augmented variables
        """
        logP = 0.
        for n in self.augment_names:
            logP += stats.norm.logpdf(x[n])
        return logP

    def _marginalise_augment(self, x, K=50):
        """
        Marginalise out the augmented feautures
        """
        x_prime, log_J = self.rescale(x, generate_augment=False)
        x_prime = live_points_to_array(x_prime, names=self.rescaled_names)
        x_prime = np.repeat(x_prime, K, axis=0)

        x_prime[:, -self.augment_features:] = np.random.randn(
                x.size * K, self.augment_features)
        z, log_prob = self.flow.forward_and_log_prob(x_prime)
        log_prob = log_prob.reshape(K, x.size)
        log_prob_e = np.sum(stats.norm.logpdf(
            x_prime[:, -self.augment_features:]), axis=-1
            ).reshape(K, x.size)
        assert log_prob.shape == log_prob_e.shape
        log_prob = -np.log(K) + logsumexp(log_prob - log_prob_e, (0))
        return log_prob - log_J

    def log_prior(self, x):
        """
        Compute the prior probability
        """
        physical_prior = self.model.log_prior(x[self.model.names])
        augmented_prior = self.augmented_prior(x)
        # return physical_prior
        return physical_prior + augmented_prior

    def compute_weights(self, x, log_q):
        """
        Compute the weight for a given set of samples
        """
        # log_q = self._marginalise_augment(x)
        log_p = self.log_prior(x)
        x['logP'] = log_p
        x['logL'] = log_q
        log_w = log_p - log_q
        log_w -= np.nanmax(log_w)
        return log_w, log_q

    def populate(self, worst_point, N=10000, plot=True):
        """Populate a pool of latent points"""
        if self.fixed_radius:
            r = self.fixed_radius
        else:
            worst_z, worst_q = self.forward_pass(worst_point, rescale=True)
            # worst_log_q = self._marginalise_augment(worst_point)
            # worst_log_q = worst_q
            r = self.radius(worst_z)
        logger.debug(f'Populating proposal with lantent radius: {r:.5}')
        warn = True
        if not self.keep_samples or not self.indices:
            self.x = np.array([], dtype=self.x_dtype)
            z_samples = np.empty([0, self.dims])
        counter = 0
        zero_counter = 0
        while len(self.x) < N:
            while True:
                z = self.draw_latent_prior(self.dims, r=r, N=self.drawsize,
                                           fuzz=self.fuzz)
                if z.size:
                    break

            x, log_q = self.backward_pass(z, rescale=True)
            # rescale given priors used initially, need for priors
            log_w, log_q = self.compute_weights(x, log_q)
            # x = x[log_q > worst_log_q]
            # log_w = log_w[log_q > worst_log_q]
            log_u = np.log(np.random.rand(x.shape[0]))
            indices = np.where((log_w - log_u) >= 0)[0]

            if not len(indices) or (len(indices) / self.drawsize < 0.001):
                logger.warning(
                    'Rejection sampling produced almost zero samples!')
                zero_counter += 1
                if zero_counter == 10 and self.x.size < (N // 2) and \
                        self.latent_prior != 'uniform':
                    logger.warning(
                        'Proposal is too ineffcient, ''reducing radius')
                    r *= 0.99
                    logger.warning(f'New radius: {r}')
                    self.x = np.array([], dtype=self.x_dtype)
                    z_samples = np.empty([0, self.dims])
                    zero_counter = 0
                    continue
            if len(indices) / self.drawsize < 0.01:
                if warn:
                    logger.warning(
                        'Rejection sampling accepted less than 1 percent of '
                        f'samples! ({len(indices) / self.drawsize})'
                        )
                    warn = False

            # array of indices to take random draws from
            self.x = np.concatenate([self.x, x[indices]], axis=0)
            z_samples = np.concatenate([z_samples, z[indices]], axis=0)
            if counter % 10 == 0:
                logger.debug(f'Accepted {self.x.size} / {N} points')
            counter += 1

        if plot:
            plot_live_points(
                self.x,
                filename=f'{self.output}/pool_{self.populated_count}.png'
                )
        self.samples = self.x[self.model.names + ['logP', 'logL']]

        if self.check_acceptance:
            self.approx_acceptance.append(self.compute_acceptance(worst_q))
            logger.debug('Current approximate acceptance: '
                         f'{self.approx_acceptance[-1]}')
            self.evaluate_likelihoods()
            self.acceptance.append(
                self.compute_acceptance(worst_point['logL']))
            logger.debug(f'Current acceptance {self.acceptance[-1]}')
        else:
            self.samples['logL'] = np.zeros(self.samples.size,
                                            dtype=self.samples['logL'].dtype)

        self.indices = np.random.permutation(self.samples.size).tolist()
        self.populated_count += 1
        self.populated = True
        self.logger.info(
            f'Proposal populated with {len(self.indices)} samples')




import numpy as np
import pytest

import nessai.livepoint as lp


@pytest.fixture
def live_point():
    return np.array([(1., 2., 3., 0., 0.)],
                    dtype=[('x', lp.DEFAULT_FLOAT_DTYPE),
                           ('y', lp.DEFAULT_FLOAT_DTYPE),
                           ('z', lp.DEFAULT_FLOAT_DTYPE),
                           ('logP', lp.DEFAULT_FLOAT_DTYPE),
                           ('logL', lp.LOGL_DTYPE)])


@pytest.fixture
def live_points():
    return np.array([(1., 2., 3., 0., 0.),
                     (4., 5., 6., 0., 0.)],
                    dtype=[('x', lp.DEFAULT_FLOAT_DTYPE),
                           ('y', lp.DEFAULT_FLOAT_DTYPE),
                           ('z', lp.DEFAULT_FLOAT_DTYPE),
                           ('logP', lp.DEFAULT_FLOAT_DTYPE),
                           ('logL', lp.LOGL_DTYPE)])


@pytest.fixture
def empty_live_point():
    return np.empty(0, dtype=[('x', lp.DEFAULT_FLOAT_DTYPE),
                              ('y', lp.DEFAULT_FLOAT_DTYPE),
                              ('z', lp.DEFAULT_FLOAT_DTYPE),
                              ('logP', lp.DEFAULT_FLOAT_DTYPE),
                              ('logL', lp.LOGL_DTYPE)])


def test_parameters_to_live_point(live_point):
    """
    Test function that produces a single live point given the parameter
    values for the live point as a live or an array
    """
    x = lp.parameters_to_live_point([1., 2., 3.], ['x', 'y', 'z'])
    np.testing.assert_array_equal(live_point, x)


def test_empty_parameters_to_live_point(empty_live_point):
    """
    Test behaviour when an empty parameter is parsed
    """
    np.testing.assert_array_equal(
        empty_live_point,
        lp.parameters_to_live_point([], ['x', 'y', 'z']))


def test_numpy_array_to_live_point(live_point):
    """
    Test the fuction the produces an array of live points given numpy array
    of shape [# dimensions]
    """
    array = np.array([1., 2., 3.])
    x = lp.numpy_array_to_live_points(array, names=['x', 'y', 'z'])
    np.testing.assert_array_equal(live_point, x)


def test_numpy_array_multiple_to_live_points(live_points):
    """
    Test the fuction the produces an array of live points given numpy array
    of shape [# point, # dimensions]
    """
    array = np.array([[1., 2., 3.], [4., 5., 6.]])
    x = lp.numpy_array_to_live_points(array, names=['x', 'y', 'z'])
    np.testing.assert_array_equal(live_points, x)


def test_empty_numpy_array_to_live_points(empty_live_point):
    """
    Test the fuction the produces an array of live points given an empty
    numpy array
    """
    np.testing.assert_array_equal(
        empty_live_point,
        lp.numpy_array_to_live_points(np.array([]), names=['x', 'y', 'z']))


def test_dict_to_live_point(live_point):
    """
    Test the function that converts a dictionary with a signel live point to
    a live point array
    """
    d = {'x': 1, 'y': 2, 'z': 3}
    x = lp.dict_to_live_points(d)
    np.testing.assert_array_equal(live_point, x)


def test_dict_multiple_to_live_points(live_points):
    """
    Test the function that converts a dictionary with multiple live points to
    a live point array
    """
    d = {'x': [1, 4], 'y': [2, 5], 'z': [3, 6]}
    x = lp.dict_to_live_points(d)
    np.testing.assert_array_equal(live_points, x)


def test_empty_dict_to_live_points(empty_live_point):
    """
    Test the function that converts a dictionary with empty lists to
    a live point array
    """
    np.testing.assert_array_equal(
        empty_live_point,
        lp.dict_to_live_points({'x': [], 'y': [], 'z': []}))


def test_live_point_to_numpy_array(live_point):
    """
    Test coversion from a live point to an unstructured numpy array
    """
    np.testing.assert_array_equal(
        np.array([[1, 2, 3, 0, 0]]),
        lp.live_points_to_array(live_point))


def test_live_point_to_numpy_array_with_names(live_point):
    """
    Test coversion from a live point to an unstructured numpy array with
    specific fields
    """
    np.testing.assert_array_equal(
        np.array([[1, 3, 0]]),
        lp.live_points_to_array(live_point, names=['x', 'z', 'logP']))


def test_live_point_to_dict(live_point):
    """
    Test conversion of a live point to a dictionary
    """
    d = {'x': 1., 'y': 2., 'z': 3., 'logP': 0., 'logL': 0.}
    assert d == lp.live_points_to_dict(live_point)


def test_live_point_to_dict_with_names(live_point):
    """
    Test conversion of a live point to a dictionary
    """
    d = {'x': 1., 'z': 3., 'logP': 0.}
    assert d == lp.live_points_to_dict(live_point, names=['x', 'z', 'logP'])


def test_multiple_live_points_to_dict(live_points):
    """
    Test conversion of multiple_live points to a dictionary
    """
    d = {'x': [1, 4], 'y': [2, 5], 'z': [3, 6], 'logP': [0, 0], 'logL': [0, 0]}
    d_out = lp.live_points_to_dict(live_points)
    assert list(d.keys()) == list(d_out.keys())
    np.testing.assert_array_equal(list(d.values()), list(d_out.values()))



import numpy as np
from numpy.testing import assert_allclose, assert_equal
import pytest

import nessai.gw.utils as utils

from conftest import requires_dependency


@pytest.mark.parametrize("r, s, zero", [((0, np.pi), 2, 'bound'),
                                        ((0, 2 * np.pi), 1, 'bound'),
                                        ((0, 1), np.pi, 'bound'),
                                        ((-np.pi, np.pi), 1, 'centre')])
@pytest.mark.parametrize('radial', [False, True])
def test_angle_to_cartesian_to_angle(r, s, zero, radial):
    """
    Test the conversion from angle to cartesian and back for a range
    and scale.
    """
    n = 1000
    t = np.random.uniform(r[0], r[1], n)
    if radial:
        radii = np.random.rand(n)
    else:
        radii = None
    cart = utils.angle_to_cartesian(t, r=radii, scale=s)
    t_out = utils.cartesian_to_angle(*cart[:2], scale=s, zero=zero)
    assert_allclose(t, t_out[0])
    if radial:
        assert_allclose(radii, t_out[1])
    assert_allclose(cart[-1], -t_out[-1])


@pytest.mark.parametrize("r, s, zero", [((0, np.pi), 2, 'bound'),
                                        ((0, 2 * np.pi), 1, 'bound'),
                                        ((0, 1), np.pi, 'bound'),
                                        ((-np.pi, np.pi), 1, 'centre')])
def test_cartesian_to_angle_to_cartesian(r, s, zero):
    """
    Test the conversion from cartesian to angle and back.
    """
    cart = np.random.randn(2, 1000)
    angles = utils.cartesian_to_angle(*cart, scale=s, zero=zero)
    cart_out = utils.angle_to_cartesian(angles[0], r=angles[1], scale=s)

    assert_allclose(cart, cart_out[:2])
    assert_allclose(angles[-1], -cart_out[-1])


@pytest.mark.parametrize("dl", [True, False])
def test_ra_dec_to_cartesian_to_ra_dec(dl):
    """
    Test the conversion from ra, dec and dl to x, y, z and back. If dl=True
    the radial component is included.
    """
    sky = np.random.uniform((0, -np.pi / 2, 100),
                            (2 * np.pi, np.pi / 2, 10000), [1000, 3]).T
    if dl:
        cart = utils.ra_dec_to_cartesian(*sky)
    else:
        cart = utils.ra_dec_to_cartesian(*sky[:2], dL=None)

    sky_out = utils.cartesian_to_ra_dec(*cart[:3])
    assert_allclose(sky[:2], sky_out[:2])
    if dl:
        assert_allclose(sky[2], sky_out[2])
    assert_allclose(cart[3], -sky_out[3])


def test_cartesian_to_ra_dec_to_cartesian():
    """
    Assert that the coversion from x, y, z, to ra, dec, dL and back
    is self consistent
    """
    cart = np.random.randn(3, 1000)
    sky = utils.cartesian_to_ra_dec(*cart)
    cart_out = utils.ra_dec_to_cartesian(*sky[:3])
    assert_allclose(cart, cart_out[:3])
    assert_allclose(sky[3], -cart_out[3])


@pytest.mark.parametrize("dl", [True, False])
def test_az_zen_to_cartesian_to_az_zen(dl):
    """
    Test the conversion from azimuth, zenith and dl to x, y, z and back.
    If dl=True the radial component if included.
    """
    sky = np.random.uniform((0, 0, 100),
                            (2 * np.pi, np.pi, 10000), [1000, 3]).T
    if dl:
        cart = utils.azimuth_zenith_to_cartesian(*sky)
    else:
        cart = utils.azimuth_zenith_to_cartesian(*sky[:2], dL=None)

    sky_out = utils.cartesian_to_azimuth_zenith(*cart[:3])
    assert_allclose(sky[:2], sky_out[:2])
    if dl:
        assert_allclose(sky[2], sky_out[2])
    assert_allclose(cart[3], -sky_out[3])


def test_cartesian_to_az_zen_to_cartesian():
    """
    Assert that the coversion from x, y, z, to azimuth, zenith, dL and back
    is self consistent
    """
    cart = np.random.randn(3, 1000)
    sky = utils.cartesian_to_azimuth_zenith(*cart)
    cart_out = utils.azimuth_zenith_to_cartesian(*sky[:3])
    assert_allclose(cart, cart_out[:3])
    assert_allclose(sky[3], -cart_out[3])


@pytest.mark.parametrize("mode", ['split', 'duplicate', 'half'])
def test_zero_one_to_cartesian(mode):
    """
    Test is correctly applied when mapping [0, 1] to
    cartesian coordinates with a given mode.
    """
    x = np.random.rand(1000)
    cart = utils.zero_one_to_cartesian(x, mode=mode)
    x_out = utils.cartesian_to_zero_one(cart[0], cart[1])

    if mode == 'duplicate':
        assert_allclose(x, x_out[0][:x.size])
        assert_allclose(x, x_out[0][x.size:])
        assert_equal([c.size for c in cart], [2 * x.size for _ in cart])
    else:
        assert_allclose(x, x_out[0])
        assert_equal([c.size for c in cart], [x.size for _ in cart])

    if mode == 'half':
        assert (cart[1] >= 0).all()
    else:
        assert ((cart[0] < 0) & (cart[1] < 0)).any()
        assert ((cart[0] > 0) & (cart[1] > 0)).any()
        assert ((cart[0] < 0) & (cart[1] > 0)).any()
        assert ((cart[0] > 0) & (cart[1] < 0)).any()


def test_zero_one_to_cartesain_incorrect_mode():
    """
    Test to ensure that an incorrect mode raises an error
    """
    x = np.random.rand(1000)
    with pytest.raises(RuntimeError) as excinfo:
        utils.zero_one_to_cartesian(x, mode='roar')
    assert 'Unknown mode' in str(excinfo.value)


def test_cartesian_to_zero_one():
    """
    Test to ensure values are mapped to [0, 1]
    """
    cart = np.random.randn(2, 1000)
    x, _, _ = utils.cartesian_to_zero_one(*cart)
    assert np.logical_and(x >= 0, x <= 1).all()


@requires_dependency('lal')
def test_precessing_parameters():
    """
    Test to ensure spin coversions are invertible
    """
    n = 1000
    theta_jn = np.arccos(np.random.uniform(-1, 1, n))
    phi_jl = np.random.uniform(0, 2 * np.pi, n)
    phi_12 = np.random.uniform(0, 2 * np.pi, n)
    theta_1 = np.arccos(np.random.uniform(-1, 1, n))
    theta_2 = np.arccos(np.random.uniform(-1, 1, n))
    a_1 = np.random.uniform(0, 0.99, n)
    a_2 = np.random.uniform(0, 0.99, n)
    m1 = 36.0
    m2 = 29.0
    phase = 0.0
    f_ref = 50.0

    array_in = (theta_jn, phi_jl, theta_1, theta_2, phi_12, a_1, a_2)

    array_inter = utils.transform_from_precessing_parameters(
        *array_in, m1, m2, f_ref, phase)

    array_out = utils.transform_to_precessing_parameters(
        *array_inter[:-1], m1, m2, f_ref, phase)

    np.testing.assert_array_almost_equal(array_in, array_out[:-1])
    np.testing.assert_array_almost_equal(array_inter[-1], -array_out[-1])
